unit Alpha;

{ -----------------------------------------------------------------------------}
{ A simple Alphabet Button Pad component for Delphi.                           }
{ Copyright 1995, Rob Marsh, S.J. All Rights Reserved.                         }
{ This component can be freely used and distributed in commercial and private  }
{ environments, provided this notice is not modified in any way.               }
{ -----------------------------------------------------------------------------}
{ Feel free to contact me if you have any questions, comments or suggestions   }
{ at RobMarsh@aol.com.                                                         }
{ -----------------------------------------------------------------------------}
{ Date last modified:  04/04/95                                                }
{ -----------------------------------------------------------------------------}

{ -----------------------------------------------------------------------------}
{ TAlphaPad v1.00                                                              }
{ -----------------------------------------------------------------------------}
{ Description:                                                                 }
{   A Button Pad to permit selection of letters of the alphabet. Might be      }
{   useful, e.g., in an address book application.                              }
{ Features:                                                                    }
{   Two orientations -- horizontal and vertical;                               }
{   Two styles       -- single and double width;                               }
{ Notes:                                                                       }
{   If the Width and Height properties of TAlphaPad do not permit an exact fit }
{   for the pad of buttons the 'end' buttons will take up the slack. This      }
{   looks bad in most cases so choose the size carefully for a good visual     }
{   appearance.                                                                }
{ -----------------------------------------------------------------------------}

interface

uses Graphics, Classes, ExtCtrls, Buttons, Messages, Controls;

type
  { The 'return' value from the pad is represented by the subrange 1..26       }
  { rather than by 'A'..'Z'.                                                   }
  TValue = 1..26;
  TOrientation = (aoHorizontal, aoVertical);
  TStyle = (asSingle, asDouble);

type
  { TAlphaPad is basically a Panel with 26 Speed Buttons. }
  TAlphaPad = class(TCustomPanel)
  private
    { The 26 Speed Buttons. }
    Pad : array [TValue] of TSpeedButton;
    FOnChange: TNotifyEvent;
    FValue: TValue;
    FOrientation: TOrientation;
    FStyle: TStyle;
    procedure PadClick(Sender: TObject);
    procedure Size(newwidth, newheight: Integer);
    procedure SizeH1(w, h: Integer);
    procedure SizeH2(w, h: Integer);
    procedure SizeV1(w, h: Integer);
    procedure SizeV2(w, h: Integer);
    procedure SetValue(Value: TValue);
    procedure SetOrientation(Value: TOrientation);
    procedure SetStyle(Value: TStyle);
  public
    constructor Create( AOwner: TComponent ); override;
  protected
    procedure Change; dynamic;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
  published
    { 'Re-publish' selected properties of TCustomPanel, ... }
    property Align;
    property Enabled;
    property Font;
    property ParentFont;
    property ParentShowHint;
    property ShowHint;
    property Visible;
    { ... and introduce some new ones -- ... }
    { ... the index of the selected button, ... }
    property Value: TValue read FValue write SetValue;
    { ... i.e. horizontal or vertical, ... }
    property Orientation: TOrientation read FOrientation write SetOrientation;
    { ... i.e. single or double 'width' --  ... }
    property Style: TStyle read FStyle write SetStyle;
    { ... and a new event. }
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

  { This procedure registers TAlphaPad with the Delphi Component Palette. }
  procedure Register;

implementation

uses WinProcs, Forms, WinTypes;

{ Create the new component ... }
constructor TAlphaPad.Create( AOwner: TComponent );
var i : Integer;
begin
  { ... by first calling the panel's constructor ... }
  inherited Create(AOwner);
  { ... and then creating 26 Speed Buttons to fill it. }
  For i := 1 to 26 do
    begin
    Pad[i]:=TSpeedButton.Create(Self);
    with Pad[i] do
      begin
      { Each button needs to be 'grafted' into the parent panel, ... }
      Parent := TAlphaPad(Self);
      Align:= alNone;
      { ... set to function as one with the whole group ... }
      GroupIndex := 1;
      { ... so that only one button is down at any time, ... }
      AllowAllUp:=False;
      { ... and given it's 'letter.' }
      Caption := Chr(Ord('A')+i-1);
      { When the individual button is clicked the pad needs to respond. }
      OnClick:=PadClick;
      end;
    end;
  { By default the pad is horizontal, two rows deep, and starts with 'A' down. }
  FOrientation:=aoHorizontal;
  FStyle:=asDouble;
  FValue:=1;
  Pad[FValue].Down:=true;
  { N.B. the default size is a multiple of 13x2 -- it's the only way to get    }
  { evenly-sized buttons.                                                      }
  Size(260,40);
end;

{ 'Size' simply checks the style and orientation and calls the appropriate     }
{ procedure to draw the pad at it's new size.                                  }
procedure TAlphaPad.Size(newwidth, newheight: Integer);
var i,j : Integer;
begin
  if FOrientation = aoHorizontal then
    if FStyle = asSingle then
       SizeH1(newwidth,newheight)
    else
       SizeH2(newwidth,newheight)
  else
    if FStyle = asSingle then
       SizeV1(newwidth,newheight)
    else
       SizeV2(newwidth,newheight);
end;

{ e.g. a single, horizontal pad. }
procedure TAlphaPad.SizeH1(w, h: Integer);
var i,dw: Integer;
begin
  dw:=w div 26;
  For i := 1 to 26 do
    with Pad[i] do
      begin
      Width:=dw;
      Height:=h;
      Top:=0;
      Left:=(i-1)*Width;
    end;
  { The 'last' button takes up any extra room. }
  Pad[26].Width:=w - 25*dw;
end;

procedure TAlphaPad.SizeH2(w, h: Integer);
var i,j,dw,dh: Integer;
begin
  for j := 0 to 1 do
    begin
    dw:=w div 13;
    dh:=h div 2;
    For i := 1 to 13 do
      with Pad[j*13+i] do
        begin
        Width:=dw;
        Height:=dh;
        Top:=j*Height;
        Left:=(i-1)*Width ;
        end;
    Pad[j*13+13].Width:=w-12*dw;
    end;
end;

procedure TAlphaPad.SizeV1(w, h: Integer);
var i,dh: Integer;
begin
  dh:=h div 26;
  For i := 1 to 26 do
    with Pad[i] do
      begin
      Width:=w;
      Height:=dh;
      Top:=(i-1)*Height;
      Left:=0;
    end;
  Pad[26].Height:=h-25*dh;
end;

procedure TAlphaPad.SizeV2(w, h: Integer);
var i,j,dw,dh: Integer;
begin
  dw:=w div 2;
  dh:=h div 13;
  for j := 0 to 1 do
    begin
    For i := 1 to 13 do
      with Pad[j*13+i] do
        begin
        Width:=dw;
        Height:=dh;
        Top:=(i-1)*Height;
        Left:=j*Width ;
        end;
    Pad[j*13+13].Height:=h-12*dh;
    end;
end;

{Whenever the pad's Value property is changed the appropriate button is depressed.}
procedure TAlphaPad.SetValue(Value: TValue);
begin
  FValue:=Value;
  Pad[Value].Down:=true;
end;

{ Whenever the orientation changes the pad is re-sized. }
procedure TAlphaPad.SetOrientation(Value: TOrientation);
begin
  FOrientation:=Value;
  Size(Width,Height);
end;

{ Whenever the style changes the pad is re-sized. }
procedure TAlphaPad.SetStyle(Value: TStyle);
begin
  FStyle:=Value;
  Size(Width,Height);
end;

{ The pad needs to respond to clicks on individual buttons. }
procedure TAlphaPad.PadClick(Sender: TObject);
var ButtonValue: TValue;
begin
  { Which button was clicked? ... }
  ButtonValue := Ord(TSpeedButton(Sender).Caption[1])+1-Ord('A');
  { Set the Value property appropriately ... }
  SetValue(ButtonValue);
  { ... and trigger the new OnChange event. }
  Change;
end;

procedure TAlphaPad.Change;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

{ The pad also needs to respond to external re-size events e.g. at design time.}
procedure TAlphaPad.WMSize(var Message: TWMSize);
begin
  inherited;
  Size(Message.Width,Message.Height);
end;

{ Finally, make the component available to the Delphi Component Palette on a   }
{ page of your choice.                                                         }
procedure Register;
begin
  RegisterComponents('My Controls',[TAlphaPad])
end;

end.
