{$I WINMACRO.INC}
{$M 1024,1024}
{$C FIXED PRELOAD PERMANENT}
{
WINMACRO keyboard/mouse/code macro engine DLL for Windows

Written by Ken Henderson.

Copyright (c) 1994 Ken Henderson.  All Rights Reserved.

Last modification: 11-29-94

See the WMTypes unit for details relating to data structures used by the macro
engin.  See WMImport for exported function declarations.

This DLL provides a complete macro interface for Windows programs.  Macro playback
can be triggered by keyboard or mouse events or by program code.  Likewise, once a
macro event has been triggered, it can playback keyboard events, mouse events or call
program code.  Macros are known by the type of event they playback and are identified
as either keyboard macros or code macros.  Keyboard macros are macros in the
traditional sense:  they playback recorded keyboard (and mouse) events.  Rather than
playing back keyboard and mouse messages, code macros execute callback procedures
when hotkeys are pressed.  In conjunction with the ability to limit macro playback
to the current task, code macros provide the ability to establish accelerator keys
at runtime rather than only at design time.  That is, you don't have to have
any "macro-centric" operation to perform using the macro engine; you can use it
as a simple keyboard/mouse command processor, installing each of the routines
you'd like triggered by a given key as a code macro.

Macro hotkeys aren't limited to just "special" keys like function keys or those
involving shift key (the Control,Alt and Shift keys) combinations.  They can be
"normal" keys, as well.  For example, you could use Winmacro to redefine the tilde
key, if you'd like.  Hotkeys aren't limited to just keys, either.  They can also
be mouse buttons or combinations of keys and mouse buttons.  For example, a hotkey
could be setup as control plus the left mouse button.  Then, when the control key
was held down while the left mouse button was clicked, the associated macro would
be executed.

Using Winmacro

In order to use Winmacro, you need to call InstallHook.  InstallHook allocates a
structure containing all of the information the macro engine requires.  See the
function declaration for more information.

Once InstallHook has been called, RecordingOn (or ToggleRecording) can be called
to turn macro recording on.  All keyboard (and, optionally, mouse) events will
be captured until RecordingOff is called or the MK_TOGGLERECORD key is pressed.
All recording goes by default to the "scrap" macro -- a work macro allocated by
InstallHook.  Note that you do not need to call RecordingOn/Off if you are installing
a code macro.

After a macro has been recorded to the scrap macro, InstallMacro can be called
to copy it to a new macro record.  InstallMacro is also called to allocate code macros.
See the function declaration for more information.

Once a keyboard macro has been allocated, pressing its hotkey (which can consist of mouse
events, as mentioned above) causes the events recorded to it to be played
back.  For code macros, the callback function whose address was passed in to the
InstallMacro routine is executed.

The number of available macros and the maximum length of each is determined by the
parameters passed into the InstallHook routine.

Once macro processing is finished, UninstallHook should be called.  It reverses the
work of the InstallHook routine, functionally shutting down the macro engine.

Coding Techniques

Since exceptions may be generated by the runtime library, it was imperative that
the entry points into this DLL trapped them.  You will therefore find an "open"
try...except block around every entry point.  By "open", I mean that the On ... Do
line in the except section traps _every_ exception.  Failing to do this would cause
the DLL to abort with an error, taking the calling program with it, in the event of
an exception.  This is due to the fact that exceptions are _not_ communicated across
the DLL/caller boundary.

I have stayed away from absolute variables to maintain some degree of portability.
Though it requires more program code, I've typecasted rather than used absolute
variables to convert variables of the same size between types.

All types and most constants used by this library are contained in the WMTypes
unit.  The exceptions to this are typed constants.  Since typed constants are
actually variables and are stored in a program's data segment, .EXE's and .DLL's
that use the same typed constant or variable out of a unit each get their own "copy" of
the variable.  There's therefore no reason to attempt to share variables this
way.  All actual variables, then, including typed constants that could possibly change
at run-time are declared here, in the library itself.

Initially, I wrote both Entry and Exit code for this library.  I was forced to move this
code to procedures, instead, due to problems with Turbo Debugger when exiting apps
that use this library.  Everytime I exited my program, returned to TDW, then exited to
Windows, I'd get thrown all the way out of Windows.  And, yes, I had stack checking
turned off in the Exit code.  Moving the Entry and Exit code to procedures remedied this
problem.  I've noticed that the DLL that ships with RECORDER.EXE is organized in the
same fashion:  it has both a separate entry function and an exit function.

Our entry code installs the mouse and keyboard handlers, registers our window class
and performs general initialization.  The Exit code uninstalls these items and also
attempts to deallocate any macro contexts that might still be allocated, though there
should never be any.

}
library Winmacro;
{$D WINMACRO.DLL Macro Engine for Windows.  Copyright (c) 1994 by Ken Henderson.  All Rights Reserved.}
{$R WM.RES}
uses WinTypes,
  Messages,
  WinProcs,
  SysUtils,
  WMTypes;

const
  EmptyStr : String = '';
  EmptyPChar : PChar = '';

  {Used for reading/writing macro files}

  Version : Word = $0101; {Convention is: high order = major version number,
                           low order = minor version number}
  SignatureLen = 8;
  Signature : string[SignatureLen] = 'WINMACRO';

  {Used to flag whether the library has been initialized}

  LibraryInitialized : Boolean = false;

  {Used to store the maximum number of keys that the engine will recognize}

  MaxPossibleMKeys : Word = 0;

  {Global macro engine status variable}

  MacroState : TMacroState = MSNone;

var

  {The array that holds our pointers into the macro engine.  Up to MaxMacroContext separate
   macro contexts may be active at once.}

  MacroContexts : array[0..MaxMacroContext] of pTMacroContext;

  {Global handles for the Keyboard, Mouse, Record, and Playback hooks}

  KeyboardHookHandle, MouseHookHandle, RecordHookHandle, PlaybackHookHandle : HHOOK;

  {Handle for our message processing Window.  Winmacro communicates with itself
  via custom WM_USER+ messages.  This window handles those messages.}

  MacroMsgWindow : HWND;

  {Pointers to the global ScrapMacro (to which all recording goes by default) and
   the CurrentMacro (which is used during playback).}

  ScrapMacroRec, CurrentMacroRec : pTMacroRec;

  {PlayingBackSlow is used to allow the PlaybackHook procedure to know that it has
  just returned some non-zero amount of time to the system in order to cause
  the system to wait before processing the next playback message.  It needs this
  flag to know to return 0 the next time it is called and asked for the same
  message.  This oddity is due to the strangeness of the Windows journaling API.}

  PlayingBackSlow : Boolean;

  {SaveKeyboardState allows us to take a snapshot of the keyboard just prior
  to forcing the state of the shift keys.  Forcing their state is necessary
  so that keyboard macros will play back correctly.  SaveKeyboardState saves
  the keyboard just prior to macro playback so that we may restore it afterward.}

  SaveKeyboardState : TKeyboardState;

  {Position stores the current position in the macro being played back.}

  Position : Word;

  {SaveToggles is used to store the status of the toggle keys (CapsLock, NumLock
  and ScrollLock) before they are reset by the macro playback routines.  The playback
  routines set them because they must be set to their respective states at the
  time the macro about to be played back was recorded.  Without this precaution,
  there is no guarantee that the macro will play back properly.}

  SaveToggles : Byte;

  {InitialMousePosition is used to store the initial position of the mouse pointer
  so that macros can be recorded and played relative to it.}

  InitialMousePosition : TPoint;

  {The following are global versions of the parms passed in from the caller}

  CallingWindowHook : HWND;
  MaxMacroNum : Word;
  MaxMacroLen : Word;
  FastPlayback : Boolean;
  MouseBase : TMouseBase;
  CancelChecking : Boolean;

  {TaskScope is set to either 0 or to the current task handle based on whether
   the SystemWide parameter passed into InstallHook was true or false, respectively.}

  TaskScope : THandle;

  {LoMsgBound and HiMsgBound are "exclusionary" mouse message boundary markers.  That is,
  rather than define the range of mouse messages that the engine will record, they
  define the range that it will not record.  I did it this way because there are
  messages on either side of WM_MOUSEFIRST that need to be recorded regardless of
  InstallHook's RecordMouse parameter.  Let's say, for example, that RecordMouse is
  set to ClicksAndDrags:  WM_KEYFIRST..WM_KEYLAST still need to be recorded.  As do
  all mouse messages with the exception of WM_MOUSEMOVE.  If RecordMouse is passed
  in as Ignore, we still want all messages passed to the RecordPrcc with the
  exception of those between WM_MOUSEFIRST and WM_MOUSELAST.  If RecordMouse is
  passed as All, we set both the LoMsgBound and HiMsgBound to WM_NULL so that no
  messages are actually excluded from recording.}

  LoMsgBound, HiMsgBound : Word;

 {PossibleMKeysLinear is a "linear" (single dimensional) array that stores all the
  possible keys that the engine will recognize.  Note that this array isn't actually
  used by the engine to determine whether it "knows" of a given pressed key.  Rather,
  it is provided for the sole purpose of allowing applications that use this library
  to easily provide lists of available hot keys.  Note also that the appearance of a
  given key combination in this list does not mean that _Windows_ will recognize the
  key when pressed; there are key combinations that, while valid, go unnoticed by
  Windows.  I have attempted to eliminate key combinations from the list that Windows
  doesn't recognize, but I might have missed some.  The best way to determine whether
  Windows recognizes a given combination is to drop a TMacroKeyComboBox component
  on to a form and attempt, run the app, click the box and press the key combination.
  If the box displays a textual name for the key you pressed, you're in luck.  If not,
  Windows probably doesn't recognize the key combination you pressed.

  An example of an unrecognized combination is Ctrl+Alt+F11.  Ctrl+Alt+F12 is recognized.
  So is Ctrl+Alt+F10, but not Ctrl+Alt+F11.  Another special case is the Cancel key,
  commonly known as the Ctrl+Break key.  Windows returns a special virtual
  key code when Control is held down while the Pause key is pressed.  Rather than
  receiving VK_PAUSE, and then being able to detect that the Control key is
  currently being pressed, we receive VK_CANCEL.  This is true for any key combination
  involving both the Pause key and the Control key: we always receive VK_CANCEL from
  Windows, though we can then check to see what shift keys are also depressed.  Those
  combinations of the Pause key and shift keys other than Control, _do not_ yield
  VK_CANCEL, but rather VK_PAUSE.  So Shift+Pause is represented as Shift+Pause,
  while Control+Shift+Pause is represented as Control+Shift+Cancel.

  I describe the array here as "linear" because the Windows virtual key codes and all
  of their possible shifted variations logically belong in a two-dimentional array.
  In considering how this array will be used, though, it seemed to me that the most
  popular use of this array would be to populate list boxes and so forth --
  single-dimensional lists of some kind -- hence my use of a single dimensional array.}

 PossibleMKeysLinear : array[0..(succ(high(byte))*succ(byte(PSCtrlShiftAlt)))] of Word;

 {Buffer for working with PChar's}

const
  WorkBufLen = 40;
var
  WorkBuf : array[0..WorkBufLen] of Char;

{Forward pointer to our message window's window proc.}

function MacWinProc(Window: HWnd; Message, WParam: Word;
  LParam: Longint): Longint; export; forward;

const

 {Macro message window class}

  MacClass: TWndClass = (
    style: 0;
    lpfnWndProc: @MacWinProc;
    cbClsExtra: 0;
    cbWndExtra: 0;
    hInstance: 0;
    hIcon: 0;
    hCursor: 0;
    hbrBackground: 0;
    lpszMenuName: nil;
    lpszClassName: 'MacroWin');

  {Default keys -- these can be manipulated using the appropriate Get/SetXXXXKey routines.}
  {$I WMDFAULT.INC}
  MK_TOGGLERECORD : Word = DEFAULTMK_TOGGLERECORD;
  MK_PAUSE : Word = DEFAULTMK_PAUSE;
  MK_PLAYBACKSCRAP: Word = DEFAULTMK_PLAYBACKSCRAP;
  MK_CANCEL : Word = DEFAULTMK_CANCEL;

  {Procedures for working with bit tables}

procedure SetBit(var BitTable : Byte; BitMask : Byte);
{  inline($58/$5F/$07/$26/$08/$05);}
begin
  BitTable:=BitTable or Bitmask;
end;

procedure ClearBit(var BitTable : Byte; BitMask : Byte);
{  inline($58/$5F/$07/$F6/$D0/$26/$20/$05);}
begin
  BitTable:=BitTable and not(BitMask);
end;

function BitSet(BitTable, BitMask : Byte) : Boolean;
{  inline($5A/$58/$20/$D0/$74/$02/$B0/$01);}
begin
  Result:=Boolean(BitTable and BitMask);
end;

{Generic routines to display messages}

function LoadMsg(MsgNum: Word): string; export;
begin
  Result[0] := Char(LoadString(HInstance, MsgNum, @Result[1], 254));
end;

procedure DisplayMessage(const MsgNum : Word);

var
  Buf : array[0..sizeof(String)] of Char;
begin
  MessageBox(0,StrPCopy(Buf,LoadMsg(MsgNum)),EmptyPChar,MB_TASKMODAL);
end;

{"Insulated" GetMem and FreeMem.  If enough memory for a request is not available,
 pointer is set to nil.  Also ensures that freed pointers are set to nil.}

function GetMemCheck(var InPtr; Size : Word) : Boolean;
begin
  try
  Result := true;
  If (MaxAvail<Size) then begin
     DisplayMessage(MGOutOfMemory);
     Pointer(InPtr):=nil;
     Result:=false;
  end else GetMem(Pointer(InPtr),Size);
  except
  On Exception Do begin
                    Result:=false;
                    Pointer(InPtr):=nil;
                  end;

  end;
end;

function FreeMemCheck(var InPtr; Size : Word) : Boolean;
begin
  Result := true;
  try
  FreeMem(Pointer(InPtr),Size);
  except
  On Exception Do Result:=false;
  end;
  Pointer(InPtr):=nil;
end;

{Used to calculate the memory size of a macro given its length in messages.  Note
 the use of the TMHDR type to calculate the size of all the fields in the struct
 but the message array.  Additional fields could be added to the macro header
 struct without invalidating this calculation.}

function CalcMacroSize(MacroEnd : Word) : Longint;
begin
  Result:=sizeof(TMHDR)+(Succ(MacroEnd)*sizeof(TEventMsg));
end;

{Checks to see if the macro passed in is the ScrapMacro.  The ScrapMacro isn't
valid in certain situations (e.g., deallocating it while an engine is active), so
checking for its presence is important.}

function IsScrapMacro(const pInMacro : pTMacroRec) : Boolean;
begin
  Result :=BitSet(pInMacro^.MHDR.Toggles,ScrapMacroStatusBit);
end;

{The base, primative macro creation routine.  It allocates a new macro structure,
then copies another macro (usually the ScrapMacro) to it.  It doesn't copy the Name,
CodeHook or MKey of the source macro, that is left to the caller.  It would make little
sense to copy these fields, since they are macro specific.  Having two macros with
the same MKey would cause one of them to never be executed via its MKey since the
macro that was allocated first would always be located by the engine's search
routines.}

function AllocateMacro(var pInMacro : pTMacroRec; const MacroRec : pTMacroRec) : Boolean;
var
  Size : Word;
begin
  Result:=false;
  Size:=CalcMacroSize(MacroRec^.MHDR.MacroEnd);
  If not GetMemCheck(pInMacro,Size) then Exit;
  If (@MacroRec^.MHDR.CodeHook=nil) then Move(MacroRec^.Msgs,pInMacro^.Msgs,Size-sizeof(TMHDR));
  Move(MacroRec^.MHDR,pInMacro^.MHDR,sizeof(pInMacro^.MHDR));
  With pInMacro^.MHDR do begin
    MKey:=0;
    {Since we're copying from the ScrapMacro, make sure the new macro is not mistaken for it.}
    ClearBit(Toggles,ScrapMacroStatusBit);
    Name:=nil;
    CodeHook:=nil;
  end;
  Result:=true;
end;

{The primative routine for deallocating macros.  It does not check to see
whether the macro it is deallocating is the ScrapMacro, therefore it is used to
deallocate the ScrapMacro when the engine is unloaded from memory.}

function DeallocateMacroPrim(var pInMacro : pTMacroRec) : Boolean;
begin
  With PInMacro^.MHDR do begin
    StrDispose(Name); {Get rid of the name buffer first}
    Name:=nil;
  end;
  Result:=FreeMemCheck(pInMacro,CalcMacroSize(pInMacro^.MHDR.MacroEnd));
end;

{Checks to see whether the macro passed in is the scrap macro, and, if so,
displays and error message.}

function CheckScrapMacro(pInMacro : pTMacroRec) : Boolean;
begin
  Result:=not(IsScrapMacro(pInMacro));
  If (not Result) then DisplayMessage(MGScrapInvalid);
end;

{Higher level macro deallocation routine}

function DeallocateMacro(var pInMacro : pTMacroRec) : Boolean;
begin
  Result:=false;

  If (pInMacro<>nil) then begin

    If (not CheckScrapMacro(pInMacro)) then Exit;
    Result:= DeallocateMacroPrim(pInMacro);
  end else Result:=true;
  {Don't error if a nil pointer is passed in}
end;

{Used to destroy then recreate a macro with a given MKey.}

function ReallocateMacro(var pInMacro : pTMacroRec; const MacroRec : pTMacroRec) : Boolean;
begin
  Result:=((DeallocateMacro(pInMacro)) and (AllocateMacro(pInMacro,MacroRec)));
end;

{Used to ensure that the macro context handle that has been passed in is valid.}

function CheckMacroHandle(const MH : pTMacroContext) : Boolean;
begin
  Result:=(MH<>nil);
  If not(Result) then DisplayMessage(MGInvalidHandle);
end;

{Deallocates all the user-defined macros in specified context in one fell swoop.}

function DeallocateMacros(const MH : pTMacroContext) : Boolean;
var
  Mcnt : Word;
begin
  Result:=false;
  If CheckMacroHandle(MH) then with MH^ do begin
    For Mcnt:=0 to pred(MaxMacroNum) do if not(DeallocateMacro(Macros^[Mcnt])) then
      Exit;
  end;
  Result:=true;
end;

{Used to manage the shift key status bits stored in the high order Bytes of MKeys.
 Does not require that the macro engine be initialized.}

procedure SetShiftState(var MKey : Word; const Key : ShiftKey; const State : Boolean);
begin
  If State then SetBit(WBytes(MKey)[1],ShiftStateOn[Key])
  else ClearBit(WBytes(MKey)[1],ShiftStateOn[Key]);
end;

function GetShiftState(const MKey : Word; const Key : ShiftKey) : Boolean;
begin
  Result := BitSet(WBytes(MKey)[1],ShiftStateOn[Key]);
end;

{Used to work with individual shift states directly.  Does not require that the
macro engine be initialized.}

function GetKeyCtrl(const MKey : Word) : Boolean; export;
begin
  Result:=GetShiftState(MKey, SKCtrl);
end;

procedure SetKeyCtrl(var MKey : Word; const CtrlState : Boolean); export;
begin
  SetShiftState(MKey, SKCtrl, CtrlState);
end;

function GetKeyAlt(const MKey : Word) : Boolean; export;
begin
  Result:=GetShiftState(MKey, SKAlt);
end;

procedure SetKeyAlt(var MKey : Word; const AltState : Boolean); export;
begin
  SetShiftState(MKey, SKAlt, AltState);
end;

function GetKeyShift(const MKey : Word) : Boolean; export;
begin
  Result:=GetShiftState(MKey, SKShift);
end;

procedure SetKeyShift(var MKey : Word; const ShiftState : Boolean); export;
begin
  SetShiftState(MKey, SKShift, ShiftState);
end;

function BuildPossibleMKeys : Integer;
const
  MK_CTRLALTF11=$37A;
  MK_CTRLALTDEL=$32E;
  {Control+Alt+Shift+Delete hangs machines, can't use it either}
  MK_CTRLALTSHIFTDEL=$72E;
var
  MKey : Word;
  MCnt : Byte;
  Scnt : PossibleShifts;
begin
  Result :=0;
  For Mcnt:=1 to VK_LAST do begin
    MKey:=Mcnt;
    If (MKey in VKeys) then
      For Scnt:=PSNone to PSCtrlShiftAlt do begin
        WBytes(MKey)[1]:=PossibleShiftStates[Scnt];
        {If Pause is pressed while Control is held down, VK_CANCEL is returned, not VK_PAUSE}
        If (Mcnt=VK_PAUSE) and (ByteBool(WBytes(MKey)[1] and CtrlOn)) then continue;
        {Control+Alt+F11 isn't returned by Windows, don't add it to the list};
        If (MKey=MK_CTRLALTF11) then continue;
        {If Pause is pressed without Control down, VK_PAUSE is returned, not VK_CANCEL}
        If (MCnt=VK_CANCEL) and ((WBytes(MKey)[1]<>0) and not(ByteBool(WBytes(MKey)[1] and CtrlOn))) then continue;
        {Conrol+VK_CANCEL is redundant, eliminate it from the list}
        If (MCnt=VK_CANCEL) and (WBytes(MKey)[1]=CtrlOn) then continue;
	 {Control+Alt+Delete is intercepted by Windows, can't use it}
        If (MKey=MK_CTRLALTDEL) then continue;
	 {Control+Alt+Shift+Delete hangs machines, can't use it either}
        If (MKey=MK_CTRLALTSHIFTDEL) then continue;
        PossibleMKeysLinear[Result]:=MKey;
        Inc(Result);
      end;
  end;
end;

function GetPossibleMKeyIndex(const MKey : Word) : Integer; export;
var
  Mcnt : Integer;
begin
  Result:=-1;
  try
  If (MaxPossibleMKeys=0) then MaxPossibleMKeys:=BuildPossibleMKeys;
  For Mcnt:=0 to pred(MaxPossibleMKeys) do
    If (PossibleMKeysLinear[Mcnt]=MKey) then begin
      Result:=Mcnt;
      exit;
    end;
  except
  On Exception do begin
    Result:=-1;
    DisplayMessage(MGUnknownError);
  end;
  end;
end;

function GetPossibleMKey(const Index : Word) : Word; export;
begin
  try
  If (MaxPossibleMKeys=0) then MaxPossibleMKeys:=BuildPossibleMKeys;
  Result:=PossibleMKeysLinear[Index];
  except
  On Exception do begin
    Result:=0;
    DisplayMessage(MGUnknownError);
  end;
  end;
end;

function GetMaxPossibleMKeys : Word; export;
begin
  try
  If (MaxPossibleMKeys=0) then MaxPossibleMKeys:=BuildPossibleMKeys;
  Result:=MaxPossibleMKeys;
  except
  On Exception do begin
    Result:=0;
    DisplayMessage(MGUnknownError);
  end;
  end;
end;

{Looks up a macro pointer in the specified macro context using its MKey.}

function GetMacroByKey(const MH : pTMacroContext; const MKey : Word) : pTMacroRec; export;
var
  Mcnt : Word;
begin
  try
  Result:=nil;
  If CheckMacroHandle(MH) then With MH^ do begin
    Result:=nil;
    If (MKey=ScrapMacroRec^.MHDR.MKey) then Result:=ScrapMacroRec
    else begin
      for Mcnt:=0 to pred(MaxMacroNum) do begin
        if (Macros^[Mcnt]<>nil) then begin
          If (Macros^[Mcnt]^.MHDR.MKey=MKey) then begin
            Result:=Macros^[Mcnt];
            Exit;
          end;
        end;
      end;
    end;
  end;
  except
  On Exception Do begin
    DisplayMessage(MGUnknownError);
    Result:=nil;
  end;
  end;
end;

{Looks up a macro pointer in the specified context using its index.}

function GetMacroByIndex(const MH : pTMacroContext; const Index : Integer) : pTMacroRec; export;
begin
  try
  Result:=nil;
  If CheckMacroHandle(MH) then With MH^ do begin
    If (Index=-1) then Result:=ScrapMacroRec
    else Result := Macros^[Index];
  end;
  except
  On Exception Do begin
    DisplayMessage(MGUnknownError);
    Result:=nil;
  end;
  end;
end;

{Gets/sets the MKey associated with a macro.}

function GetMacroKey(const pInMacro : pTMacroRec) : Word; export;
begin
  try
  If (pInMacro<>nil) then Result:=pInMacro^.MHDR.MKey
  else Result:=0;
  except
  On Exception Do begin
    DisplayMessage(MGUnknownError);
    Result:=MRSevereError;
  end;
  end;
end;

function GetCurrentMacroContext(const MKey : Word) : pTMacroContext; forward;

procedure SetMacroKey(const pInMacro : pTMacroRec; const MKey : Word); export;
var
  TempMac : pTMacroRec;
begin
  try
  If (MacroState=MSNone) then begin
    If (pInMacro<>nil) then begin
      TempMac:=GetMacroByKey(GetCurrentMacroContext(MKey),MKey);
      If (TempMac=nil) or (TempMac=pInMacro) then pInMacro^.MHDR.MKey:=MKey
      else DisplayMessage(MGKeyInUse);
    end;
  end else DisplayMessage(MGCantDoNow);
  except
  On Exception Do begin
    DisplayMessage(MGUnknownError);
  end;
  end;
end;

{Gets/sets the text name associated with a macro.}

function GetMacroName(const Macro : pTMacroRec): PChar; export;
begin
  try
  If (Macro<>nil) then begin
    If (Macro^.MHDR.Name<>nil) then Result:=Macro^.MHDR.Name
    else Result:=StrPLCopy(WorkBuf,LoadMsg(MNNullMacroName),SizeOf(WorkBuf));
  end else Result:=StrPLCopy(WorkBuf,LoadMsg(MNMacroUndefined),SizeOf(WorkBuf));
  except
  On Exception Do begin
    DisplayMessage(MGUnknownError);
    Result:=nil;
  end;
  end;
end;

function StrTrim(const InPChar : PChar) : PChar;
begin
  If (InPChar<>nil) then
    While (InPChar[StrLen(InPChar)]=' ') do
      InPChar[StrLen(InPChar)]:=#0;
  Result:=InPChar;
end;

procedure SetMacroName(const pInMacro : pTMacroRec; const Name : PChar); export;
begin
  try
  if (Name<>nil) and (boolean(StrComp(StrTrim(Name),EmptyPChar))) then begin
      If (pInMacro<>nil) then
        With pInMacro^ do begin
          StrDispose(MHDR.Name);
          MHDR.Name:=StrNew(Name);
        end;
  end else pInMacro^.MHDR.Name:=nil;
  except
  On Exception Do begin
    DisplayMessage(MGUnknownError);
  end;
  end;
end;

{Gets/sets the code hook procedure associated with a macro.  Setting a macro's
 "CodeHook" changes it to a Code Macro -- one that executes a routine, rather than
 plays back key and mouse messages, when it's MKey is pressed.}

function GetMacroCodeHook(const pInMacro : pTMacroRec) : CodeHookProc; export;
begin
  try
  If (pInMacro<>nil) then Result:=pInMacro^.MHDR.CodeHook
  else Result:=nil;
  except
  On Exception Do begin
    DisplayMessage(MGUnknownError);
    Result:=nil;
  end;
  end;
end;

procedure SetMacroCodeHook(const pInMacro : pTMacroRec; const CodeHook : CodeHookProc); export;
begin
  try
  If (MacroState=MSNone) then begin
    If (pInMacro<>nil) then pInMacro^.MHDR.CodeHook:=CodeHook;
  end else DisplayMessage(MGCantDoNow);
  except
  On Exception Do begin
    DisplayMessage(MGUnknownError);
  end;
  end;
end;

{Gets the name of a macro's hotkey by inspecting the components of its MKey field.
 Note the special handling of some keys.  Windows doesn't seem to want to
 return some key names correctly.  This routine also returns names for the mouse buttons,
 since they, too, may be used in hotkey combinations.

 MKey specifies the hotkey for which to derive a name

 MKeyText points to a buffer sufficient to hold the number of Characters specified by
 MaxTextLen.

 Returns a pointer to MKeyText on success, else nil.}

function GetKeyName(const MKey : Word; const MKeyText : PChar; const MaxTextLen : Word) : PChar; export;
const
  MaxKeyTextLen = 40;

var
  MKeyLongint : Longint;
  ScanCode, ShiftState : Byte;
  KeyText : array[0..MaxKeyTextLen] of Char;
begin
  Result:=nil;
  try
  StrCopy(MKeyText,EmptyPChar);
  ShiftState:=Hi(MKey);
  If BitSet(ShiftState,CtrlOn) then StrLCat(MKeyText,'Ctrl+',MaxTextLen);
  If BitSet(ShiftState,AltOn) then StrLCat(MKeyText,'Alt+',MaxTextLen);
  If BitSet(ShiftState,ShiftOn) then StrLCat(MKeyText,'Shift+',MaxTextLen);


  Case Lo(MKey) of
    0 : StrPCopy(KeyText,LoadMsg(KNNoname));
    1,2,4 : StrPCopy(KeyText,LoadMsg(KNNoname+lo(MKey)));
    VK_PAUSE : StrPCopy(KeyText,LoadMsg(KNPause));
    VK_CANCEL : StrPCopy(KeyText, LoadMsg(KNCancel));
    VK_DIVIDE : StrPCopy(KeyText, LoadMsg(KNDivide));
    VK_DECIMAL : StrPCopy(KeyText, LoadMsg(KNDecimal));
    else begin
      ScanCode:=mapvirtualkey(Lo(MKey),0);
      MKeyLongint:=0;
      LBytes(MKeyLongint)[2]:=ScanCode;
      If (Lo(MKey) in ExtendedVKeys) then SetBit(LBytes(MKeyLongint)[3],1);
      If (not(bytebool(GetKeyNameText(MKeyLongint,KeyText,MaxKeyTextLen)))) then
        StrPCopy(KeyText,LoadMsg(KNNoname));
    end;
  end;
  StrLCat(MKeyText,KeyText,MaxTextLen);
  Result:=MKeyText;
  except
  On Exception Do begin
    DisplayMessage(MGUnknownError);
    Result:=nil;
  end;
  end;
end;

{Get/set the various default keys.}

function GetToggleRecordKey : Word; export;
begin
  Result:=MK_TOGGLERECORD;
end;

procedure SetToggleRecordKey(const MKey : Word); export;
begin
  MK_TOGGLERECORD := MKey;
end;

function GetCancelKey : Word; export;
begin
  Result:=MK_CANCEL;
end;

function SetCancelKey(const MKey : Word) : boolean; export;
begin
  If (hi(MKey)=0) then MK_CANCEL := MKey;
  Result:=(MK_CANCEL = MKey);
  If Not(Result) then DisplayMessage(MGShiftKeysNotAllowed);
end;

function GetPauseKey : Word; export;
begin
  Result:=MK_PAUSE;
end;

function SetPauseKey(const MKey : Word) : boolean; export;
begin
  If (hi(MKey)=0) then MK_PAUSE := MKey;
  Result:=(MK_PAUSE = MKey);
  If Not(Result) then DisplayMessage(MGShiftKeysNotAllowed);
end;

{Get/SetToggleStates

These procedures use Windows' GetKeyState to get the current state of the toggle
shift keys (Capslock, Numlock, and ScrollLock) so that they may be saved/restored
during macro recording/playback.  This is necessary because the status of these
toggles affects the characters generated by Windows when macros are replayed.
That is, if Capslock was not down when the user recorded a macro that types 'ken'
into the current input focus, but is down when the macro is played back, 'KEN'
would be generated without Capslock being set to its previous state.  The Toggles
field of the macro header stores this information.

}

function GetToggleStates : Byte;
var
  TempMKey : Word;
begin
  try
  TempMKey:=0;
  SetShiftState(TempMKey, SKCapslock, BitSet(GetKeyState(VK_CAPITAL),1));
  SetShiftState(TempMKey, SKNumlock, BitSet(GetKeyState(VK_NUMLOCK),1));
  SetShiftState(TempMKey, SKScrollLock, BitSet(GetKeyState(VK_SCROLL),1));
  Result:=hi(TempMKey);
  except
  On Exception Do begin
    DisplayMessage(MGUnknownError);
    Result:=MRSevereError;
  end;
  end;
end;

procedure SetToggleStates(ShiftState: Byte);
const
  KeyDown = 1;
var
  KeyBoardState : TKeyboardState;
begin
  try
  GetKeyBoardState(KeyBoardState);
  If BitSet(ShiftState,CapslockOn) then SetBit(KeyBoardState[VK_CAPITAL],KeyDown)
  else ClearBit(KeyBoardState[VK_CAPITAL],KeyDown);
  If BitSet(ShiftState,NumlockOn) then SetBit(KeyBoardState[VK_NUMLOCK],KeyDown)
  else ClearBit(KeyBoardState[VK_NUMLOCK],KeyDown);
  If BitSet(ShiftState,ScrollLockOn) then SetBit(KeyBoardState[VK_SCROLL],KeyDown)
  else ClearBit(KeyBoardState[VK_SCROLL],KeyDown);
  SetKeyBoardState(KeyboardState);
  except
  On Exception Do begin
    DisplayMessage(MGUnknownError);
  end;
  end;
end;

{Sends a message to the CallingWindowHook's of all macro contexts.  These messages are
defined in WMTypes and are used to notify the user of state changes in the macro
engine.  This is desirable since both recording and playback -- the two main causes
of these messages -- are global in scope:  only one macro context can play or record
macros at a time.}

Procedure BroadcastMacroMsg(const Message : Word; const wparam : TUserMacroState; const lParam : Longint);
var
  Mcnt : Byte;
begin
  For mcnt:=0 to MaxMacroContext do
   If (MacroContexts[mcnt]<>nil) then
     PostMessage(MacroContexts[mcnt]^.CallingWindowHook,Message,Word(wparam),lparam);
end;

{
function InstallMacro(const MH : pTMacroContext;
                      const MKey : Word;
                      const Macro : pTMacroRec;
                      const Name : PChar;
                      const CodeHook : CodeHookProc): Byte; export;

where
  MH is the handle to macro context previously allocated by the InstallHook routine.

  MKey is the hotkey to be associated with the new macro.  MKeys consist of two
  parts:  the virtual key portion and the shift state portion.  Windows' virtual
  key constants can be used to define the virtual key portion; it is stored in the
  low order Byte.  Note that these include mouse clicks, which Winmacro recognizes,
  as well.  The virtual key portion can be assigned by simply assigning the MKey
  variable since it is stored in the low order Byte.  The shift state portion of an
  MKey tells Winmacro which shift keys must be pressed along with the aforementioned
  virtual key in order to trigger the macro.  Winmacro provides a collection of
  routines to work with this portion of an MKey.  See the SetKeyXXXX and GetKeyXXXX
  routines for more information.  Winmacro also provides a complete set of constants
  to help specify hot keys.  These are found in the WMKeys.Inc include file -- you
  need only use WMTypes (or the components in WMacro) to use them.  They follow the
  form MK_CtrlAltShiftKeyName, where Ctrl signifies that Control is is to be pressed for this
  combination, Alt indicates Alt is to be pressed and Shift indicates that Shift is
  to be pressed.  KeyName is the name of the primary key.  An example of a valid MK
  constant is MK_CTRLALTSHIFTF10, which is Control+Alt+Shift+F10.

  Macro is a pointer to a macro record containing a recorded macro.  Normally, this
  will be the scrap macro.  InstallMacro copies the messages recorded in this
  macro to the new one it creates.  Pass in a nil macro pointer in order to delete
  a macro associated with the specified MKey.

  Name is the textual name of the macro.  This can later be retrieved for detailing
  recorded macros in a listbox, for example.

  CodeHook is a pointer to the call back routine to call for code macros.  Set this
  parameter to nil for keyboard macros.  If this parameter is not nil, the engine
  issues a far call to it when the hot key is pressed.  The routine must be declared
  with the "export" directive in order to used as a CodeHook proc.

returns

  a Byte indicating the success or failure of the call.  Possible return values
  are:

  MRSuccess	 = macro has been installed
  MRNoMacroRoom = no room for additional macros (probably due to the limit imposed
                  by the MaxMacNum parameter passed into InstallHook)
  MROutOfMemory = no memory for additional macro allocations, or a possible
                  unexpected memory allocation error while attempting to
                  allocate the macro
}

function InstallMacro(const MH : pTMacroContext;
                      const MKey : Word;
                      const Macro : pTMacroRec;
                      const Name : PChar;
                      const CodeHook : CodeHookProc): Byte; export;
var
  Mcnt, FirstAvail : Integer;
begin
  try
  If (MacroState=MSNone) then begin
    Result:=MROutOfMemory;
    If CheckMacroHandle(MH) then With MH^ do begin
      FirstAvail := -1;
      for Mcnt:= 0 to pred(MaxMacroNum) do begin
          if (Macros^[Mcnt]<>nil) then begin
            If (Macros^[Mcnt]^.MHDR.MKey=MKey) then begin
              If (Macro=nil) then begin
                DeallocateMacro(Macros^[Mcnt]);
                BroadcastMacroMsg(MM_MACROASSIGNED,UMSNone,0);
              end else begin
                If Not ReallocateMacro(Macros^[Mcnt],Macro) then Exit;
                Macros^[Mcnt]^.MHDR.MKey:=MKey;
                SetMacroName(Macros^[Mcnt],Name);
                SetMacroCodeHook(Macros^[Mcnt],CodeHook);
                If (@CodeHook<>nil) then Macros^[MCnt]^.MHDR.MacroEnd:=-1;
                BroadcastMacroMsg(MM_MACROASSIGNED,UMSNone,Longint(Macros^[Mcnt]));
              end;
              Result:=MRSuccess;
              Exit;
            end;
          end else If (FirstAvail=-1) then FirstAvail:=Mcnt;
      end;

      If (FirstAvail<>-1) then begin
          If not ReallocateMacro(Macros^[FirstAvail],Macro) then Exit;
          Macros^[FirstAvail]^.MHDR.MKey:=MKey;
          SetMacroName(Macros^[FirstAvail],Name);
          SetMacroCodeHook(Macros^[FirstAvail],CodeHook);
          BroadcastMacroMsg(MM_MACROASSIGNED,UMSNone,Longint(Macros^[FirstAvail]));
          Result:=MRSuccess;

      end else begin
        DisplayMessage(MGNoMacroRoom);
        Result:=MRNoMacroRoom;
      end;

    end;
  end else DisplayMessage(MGCantDoNow);
  except
  On Exception Do begin
    DisplayMessage(MGUnknownError);
    Result:=MRSevereError;
  end;
  end;
end;

{Write the macros within the specified context to a file.  Note the use of a
 nested function to handle the checking of IOResult.  It sets the outer function's
 IORes variable, which is then returned to the caller.}

function WriteMacros(const MH : pTMacroContext; const FileName : PChar) : Byte; export;
label
  Exitpoint;
Var
  Mcnt : Word;
  NameString : MacroNameString;
  NameStringLen : Byte absolute NameString;
  MacroFile : file;
  MacroFileName : PathStr;
  IORes : integer;

function CheckIOResult(const MsgNum : Word) : Boolean;
begin
  IORes:=IOResult;
  If (IORes<>0) then DisplayMessage(MsgNum);
  Result:=(IORes=0);
end;

begin
  try
  IORes:=UnknownIOError;
  If (MacroState=MSNone) then begin
    If CheckMacroHandle(MH) then With MH^ do begin
      MacroFileName := StrPas(FileName);
      Assign(MacroFile,MacroFileName);
      Rewrite(MacroFile,1);
      If not CheckIOResult(MGCreateMacroFile) then Exit;

      BlockWrite(MacroFile,Signature,SizeOf(Signature));
      If not CheckIOResult(MGWriteMacroFile) then goto ExitPoint;

      BlockWrite(MacroFile,Version,SizeOf(Version));
      If not CheckIOResult(MGWriteMacroFile) then goto ExitPoint;

        For Mcnt:=0 to pred(MaxMacroNum) do begin
          If (Macros^[Mcnt]<>nil) then begin
            With Macros^[Mcnt]^ do BlockWrite(MacroFile,Macros^[Mcnt]^,SizeOf(MHDR)-SizeOf(MHDR.Name));
            If not CheckIOResult(MGWriteMacroFile) then goto ExitPoint;
            With Macros^[Mcnt]^.MHDR do begin
              If (Name<>nil) then NameString:=StrPas(StrTrim(Name))
              else NameString:='';
              If ((Name<>nil) and (NameString<>EmptyStr)) then begin
                 BlockWrite(MacroFile,NameString,Succ(NameStringLen));
                 If not CheckIOResult(MGWriteMacroFile) then goto ExitPoint;
              end else begin
                 {Write the length byte}
                 BlockWrite(MacroFile,EmptyStr,SizeOf(NameStringLen));
                 If not CheckIOResult(MGWriteMacroFile) then goto ExitPoint;
              end;

              BlockWrite(MacroFile,Macros^[Mcnt]^.Msgs,Succ(MacroEnd)*sizeof(TEventMsg));
              If not CheckIOResult(MGWriteMacroFile) then goto ExitPoint;
            end;
          end;
        end;
        Exitpoint:
        Close(MacroFile);
        If not CheckIOResult(MGCloseMacroFile) then Exit;
        Result:=IORes;
    end;
  end else DisplayMessage(MGCantDoNow);
  except
  On Exception Do begin
    DisplayMessage(MGUnknownError);
    Result:=MRSevereError;
  end;
  end;
end;

{Read macros from a file to the specified macro context.  Note the use of a
 nested function to handle the checking of IOResult.  It sets the outer function's
 IORes variable, which is then returned to the caller.}

function ReadMacros(const MH : pTMacroContext; const FileName : PChar; const Merge : Boolean) : Byte; export;
label
  ExitPoint;
Var
  Bytesread : Word;
  NameString : MacroNameString;
  NameStringLen : Byte absolute NameString;
  NameBuffer : array[0..sizeof(NameString)] of Char;
  MacroFile : file;
  MacroFileName : PathStr;
  IORes : integer;
  TempSignature : string[SignatureLen];
  TempVersion : Word;
  SaveMKey : Word;
  SaveName : PChar;

function CheckIOResult(const MsgNum : Word) : Boolean;
begin
  IORes:=IOResult;
  If (IORes<>0) then DisplayMessage(MsgNum);
  Result:=(IORes=0);
end;

procedure SaveScrap;
{Save the scrap macro prior to any changes}
begin
  With ScrapMacroRec^.MHDR do begin
    SaveMKey:=MKey;
    If (Name<>nil) then SaveName:=StrNew(Name)
    else SaveName:=nil;
  end;
end;

procedure RestoreScrap;
{Undo any changes to the scrap macro}
begin
  With ScrapMacroRec^.MHDR do begin
    MKey:=SaveMKey;
    StrDispose(Name);
    If (SaveName<>nil) then begin
      Name:=StrNew(SaveName);
      StrDispose(SaveName);
    end else Name:=nil;
  end;
end;

function StrAlloc(const Size : Word) : PChar;
{My own StrAlloc since the one in the Strings unit doesn't work as expected.}
var
  TempPChar : PChar;
begin
  If GetMemCheck(TempPChar, Size) then Result:=TempPChar
  else Result:=Nil;
end;

begin
  try
  Result:=UnKnownIOError;
  If (MacroState=MSNone) then begin
    If CheckMacroHandle(MH) then With MH^ do begin
      MacroFileName := StrPas(FileName);
      Assign(MacroFile,MacroFileName);
      Reset(MacroFile,1);
      If not CheckIOResult(MGOpenMacroFile) then Exit;

      BlockRead(MacroFile,TempSignature,SizeOf(Signature), Bytesread);
      If not CheckIOResult(MGReadMacroFile) then goto ExitPoint;

      If (TempSignature<>Signature) then begin
        DisplayMessage(MGInvalidMacroFile);
        goto Exitpoint;
      end;

      BlockRead(MacroFile,TempVersion,SizeOf(TempVersion), Bytesread);
      If not CheckIOResult(MGReadMacroFile) then goto ExitPoint;

      If (TempVersion>Version) then begin
        DisplayMessage(MGWrongMacroVersion);
        goto Exitpoint;
      end;

      SaveScrap;

      If (Not Merge) then DeallocateMacros(MH);

        While not eof(MacroFile) do begin
          With ScrapMacroRec^ do begin
            BlockRead(MacroFile,MHDR,SizeOf(MHDR)-SizeOf(MHDR.Name), Bytesread);
            If not CheckIOResult(MGReadMacroFile) then goto ExitPoint;

            {Read the length byte of the macro name}
            BlockRead(MacroFile,NameString[0],1,Bytesread);
            If not CheckIOResult(MGReadMacroFile) then goto ExitPoint;

            {Read the name itself}
            If (NameStringLen>0) then begin
              BlockRead(MacroFile,NameString[1],NameStringLen,Bytesread);
              If not CheckIOResult(MGReadMacroFile) then goto ExitPoint;

              StrPCopy(NameBuffer,NameString);

            end else NameBuffer[0]:=#0;

            BlockRead(MacroFile,Msgs,Succ(MHDR.MacroEnd)*sizeof(TEventMsg),Bytesread);
            If not CheckIOResult(MGReadMacroFile) then goto ExitPoint;

            With MHDR do begin
              If (InstallMacro(MH,MKey,ScrapMacroRec,NameBuffer,nil)<>0) then begin
                DisplayMessage(MGAllocateMacro);
                goto Exitpoint;
              end;
            end;
          end;
        end;
        ExitPoint:
        RestoreScrap;
        Close(MacroFile);
        If not CheckIOResult(MGCloseMacroFile) then Exit;
        Result:=IORes;
    end;
  end else DisplayMessage(MGCantDoNow);
  except
  On Exception Do begin
    DisplayMessage(MGUnknownError);
    Result:=MRSevereError;
  end;
  end;
end;

{Clears the status of the "transient" shift keys -- Control, Alt and Shift.
This is necessary to keep the shift keys used with a macro hot key,
for example, from affecting the keys the macro plays back.
}

procedure ClearShiftStates;
var
  KeyBoardState : TKeyboardState;
begin
  try
  GetKeyBoardState(KeyBoardState);
  KeyBoardState[VK_CONTROL]:=0;
  KeyBoardState[VK_SHIFT]:=0;
  KeyBoardState[VK_MENU]:=0;
  SetKeyBoardState(KeyboardState);
  except
  On Exception Do begin
    DisplayMessage(MGUnknownError);
  end;
  end;
end;

{Records the status of the "transient" shift keys -- Control, Alt and Shift.  This
is used extensively by the engine to both determine when a hotkey has been pressed
and to wait until all shifts keys are up before processing a macro command request.}

procedure GetShiftStates(var MKey : Word); export;
begin
  try
  If (not (Lo(MKey) in ShiftVKeys)) then begin {Don't modify Shift-only keys}
    SetKeyCtrl(MKey, BitSet(GetKeyState(VK_CONTROL),$80));
    SetKeyAlt(MKey, BitSet(GetKeyState(VK_MENU),$80));
    SetKeyShift(MKey, BitSet(GetKeyState(VK_SHIFT),$80));
  end;
  except
  On Exception Do begin
    DisplayMessage(MGUnknownError);
  end;
  end;
end;

{This routine uses the MKey passed into it to attempt to locate a macro context
to make current.  The first order of precedence is the current application's task
handle.  If a macro context can be located that has the same task handle as the
current application -- and also has a macro whose hotkey matches the MKey -- it
is returned without further ado.  Next in line is a sequential scan through
the available macro contexts for one with a matching MKey.  Note the setting of
the FirstNotNil variable as we scan the macro contexts.  FirstNotNil identifies the
first macro context that has been allocated in the macro context array.  If our
scan through the contexts for a matching MKey fails, we return a pointer
to the first context that is not nil, if there is one.  By setting FirstNotNil
during our scan, we avoid an extra pass to find an allocated context.}

function GetCurrentMacroContext(const MKey : Word) : pTMacroContext;
var
  CurrentTask : THandle;
  mcnt, FirstNotNil : Shortint;
begin
  Result:=nil;
  FirstNotNil:=-1;

  {First search using the current task handle}
  CurrentTask := GetCurrentTask;
  for mcnt:=0 to MaxMacroContext do
    If (MacroContexts[mcnt]<>nil) and (MacroContexts[mcnt]^.TaskHandle=CurrentTask) then
      If (MacroContexts[mcnt]^.TaskScope=0) or (CurrentTask=MacroContexts[mcnt]^.TaskScope) then begin
        Result:=MacroContexts[mcnt];
        Exit;
      end;

  {Next search using the MKey only}
  for mcnt:=0 to MaxMacroContext do
    If (MacroContexts[mcnt]<>nil) and (MacroContexts[mcnt]^.TaskScope=0) then begin
      If (FirstNotNil=-1) then FirstNotNil:=mcnt;
      If (GetMacroByKey(MacroContexts[mcnt],MKey)<>nil) then begin
        Result:=MacroContexts[Mcnt];
        Exit;
      end;
    end;

  {If unsuccessful return a pointer to the first non-nil MacroContext}
  If (FirstNotNil<>-1) then Result:=MacroContexts[FirstNotNil];
end;

{When InstallHook is called, it calls this routine to add the newly allocated
context to the context array.  If there is no room in the context array or a context
has already been allocated by the same taak, the call fails.}

function AddMacroContextEntry(const MH : pTMacroContext) : Boolean;
var
  Mcnt : Byte;
begin
  Result:=false;
  for mcnt:=0 to MaxMacroContext do
    If (MacroContexts[mcnt]<>nil) and (MacroContexts[mcnt]^.TaskHandle=MH^.TaskHandle) then begin
      Result:=false;
      DisplayMessage(MGTooManyHooks);
      Exit;
    end;
  for mcnt:=0 to MaxMacroContext do
    If (MacroContexts[mcnt]=nil) then begin
      MacroContexts[mcnt]:=MH;
      Result:=true;
      Exit;
    end;
end;

{This function returns the index value of a macro context.  It allows us to
set the element in the context array to nil after a context has been deallocated.}

function GetMacroContextIndex(const MH : pTMacroContext) : ShortInt;
Var
  Mcnt : Byte;
begin
  Result:=-1;
  for mcnt:=0 to MaxMacroContext do
    if (MacroContexts[mcnt]=MH) then begin
      Result:=mcnt;
      Exit;
    end;
end;

function NormalizeMouseMsg(Msg : Word) : Word;
begin
  {If it's a non-client mouse message, convert it to the corresponding WM_MOUSEXXXX message}
  If (Msg >= WM_NCMOUSEMOVE) and (Msg <= WM_NCMBUTTONDBLCLK) then Msg:=Msg+MOUSEMSGDELTA;

  Result:=Msg
end;

{This function serves as the keyboard processor for the macro engine.  Any key
pressed comes through this routine and is dispatched accordingly.  Keys are dispatched
by posting messages to the macro message window.  It then calls the appropriate
routines.  Windows requires that this routine return 1 when it has processed a
key, or 0 when it wants Windows to.  This routine only acts on key up messages,
never on keydown messages.  When a WM_KEYDOWN message is seen for a key that
matches the MKey of an allocated macro or of some other special key, this routine
"eats" the key by returning a 1 to Windows.  When the corresponding WM_KEYUP is seen,
the appropriate message is dispatched to the macro message window.  When a hot
key is seen, the matching context is determined and sent as part of the message
to the macro message window.  The macro whose MKey matches the one pressed
is passed as part of the message via the CurrentMacroRec field in the passed
macro context.  When the macro message window receives the MM_PLAYBACK message
containing the pointer to the macro context, it typecasts the lparam parameter
to retrieve the context pointer from the message.

The CallNextHookEx API function is called both when the Code parameter that is
passed in is < 0 and also following any processing of received messages.  The later
ensures that applications that install hooks prior to applications using this
library still function properly.

Note the use of nested procedures and functions, a distinctively "un-C" practice. }

function KeyboardHookProc(Code: Integer; wParam: Word; lParam: Longint): Longint; export;
var
   MKey : Word;
   MacroToPlay : pTMacroRec;
   MH : pTMacroContext;
   KeyIsUp : Boolean;

function KeyUp(const Lparam : Longint) : Boolean;
begin
  Result:=not(bitset(LBytes(Lparam)[3],$80));
end;

procedure HandleCancel;
begin
  Case MacroState of
  MSRecord : PostMessage(MacroMsgWindow,MM_STOPRECORD,byte(true),0);
  {Never makes it here -- left in place for clarity.}
  MSPlayback : PostMessage(MacroMsgWindow,MM_STOPPLAYBACK,0,0);
  end;
  Result:=1;
end;

procedure HandlePause;
begin
  If (KeyIsUp) then begin
    If (MacroState=MSRecord) then PostMessage(MacroMsgWindow,MM_STARTPAUSE,0,0)
    else PostMessage(MacroMsgWindow,MM_STOPPAUSE,0,0)
  end;
  Result:=1;
end;

procedure HandleRecord;
begin
  if (MacroState<>MSPlayback) {Recording, idle or paused, NOT playing back}
  and (KeyIsUp) then begin
    if (MacroState=MSNone) then PostMessage(MacroMsgWindow,MM_STARTRECORD,0,0)
    else PostMessage(MacroMsgWindow,MM_STOPRECORD,0,0);
  end;
  Result:=1;  {eat the key, even if it's a keydown message}
end;

procedure HandleOtherKeys;
begin
  MH:=GetCurrentMacroContext(MKey);
  If (MH<>nil) then begin
      if (MacroState=MSNone) then begin
        MacroToPlay:=GetMacroByKey(MH,MKey);
        if (MacroToPlay<>nil) then begin
          if (KeyIsUp) then begin
             MH^.CurrentMacroRec:=MacroToPlay;
             PostMessage(MacroMsgWindow,MM_STARTPLAYBACK,0,Longint(MH));
          end;
          Result:=1;  {eat the key, even if it's a keydown message}
        end;
      end;
  end;
end;

begin
   if (code<0) then Result:=CallNextHookEx(KeyboardHookHandle,code, wParam, lParam)
   else begin
       Result:=0;   {By default, pass the message on to Windows}

       MKey:=wparam;

       {Special provision for VK_CANCEL (Control+Break), because the system already
       takes into account that Control has been pressed with it and renders a special
       VK_ for it. (The unshifted version of the same key comes back as VK_PAUSE.)}

       If not((MK_CANCEL=VK_CANCEL) and (MKey=MK_CANCEL)) then GetShiftStates(MKey);

       KeyIsUp:=KeyUp(Lparam);
       {A question may be asked as to why I did not use a case statement
        here.  It would seem that

        Case MKey of
        VK_SHIFT..VK_MENU: ...
        MK_TOGGLERECORD: ...
        else
        end;

        would work nicely.  The problem is that the MK_XXXX constants are typed
        constants -- they can change at runtime.  This allows the user the ability
        to change them to different key combinations if desired.  The compiler
        evaluates the values on the left in a case statement at compile time
        so that a jump table can be constructed.  This is impossible with the MK_XXXX
        constants, hence the use of nested if's.}

       If ((CancelChecking) and (MKey=MK_CANCEL)) then HandleCancel
       else If (MKey = MK_PAUSE) then HandlePause
       else If (MKey = MK_TOGGLERECORD) then HandleRecord
       else HandleOtherKeys;

       If (Result=0) then Result:=CallNextHookEx(KeyboardHookHandle,code,wParam, lParam);
   end;
end;

{This function serves as the mouse processor for the macro engine.  All mouse
activity comes through this routine and is dispatched accordingly.  Mouse activity is
dispatched by posting messages to the macro message window.  It then calls the appropriate
routines.  Windows requires that this routine return 1 when it has processed a
mouse event, or 0 when it wants Windows to.  This routine only acts on button up messages,
never on buttondown messages.  The TranslateMouseMsg routine uses the MessageToVKey
typed constant array declared in WMTypes to translate mouse messages into usable
key-like events.  When a down message is seen for a button combination that
matches the MKey of an allocated macro or for some other special MKey, this routine
"eats" the message by returning a 1 to Windows.  When the corresponding up message is seen,
the appropriate message is dispatched to the macro message window.  When a hot
key button is seen, the matching context is determined and sent as part of the message
to the macro message window.  The macro whose MKey matches the one pressed
is passed as part of the message via the CurrentMacroRec field in the passed
macro context.  When the macro message window receives the MM_PLAYBACK message
containing the pointer to the macro context, it typecasts the lparam parameter
to retrieve the context pointer from the message.

The CallNextHookEx API function is called both when the Code parameter that is
passed in is < 0 and also following any processing of received messages.  The later
ensures that applications that install hooks prior to applications using this
library still function properly.

Note the use of nested procedures and functions, a distinctively "un-C" practice. }

function MouseHookProc(Code: Integer; wParam: Word; lParam: Longint): Longint; export;
var
   Msg, MKey : Word;
   MacroToPlay : pTMacroRec;
   MH : pTMacroContext;
   ButtonIsUp : Boolean;

procedure HandleCancel;
begin
  If ButtonIsUp then Case MacroState of
    MSRecord : PostMessage(MacroMsgWindow,MM_STOPRECORD,byte(true),0);
    MSPlayback : PostMessage(MacroMsgWindow,MM_STOPPLAYBACK,0,0);
  end;
  Result:=1;
end;

procedure HandlePause;
begin
  If (ButtonIsUp) then begin
     If (MacroState=MSRecord) then PostMessage(MacroMsgWindow,MM_STARTPAUSE,0,0)
     else PostMessage(MacroMsgWindow,MM_STOPPAUSE,0,0);
  end;
  Result:=1;
end;

procedure HandleRecord;
begin
  if (MacroState<>MSPlayback) and (ButtonIsUp) then begin {Recording, idle or paused, NOT playing back}
    if (MacroState=MSNone) then PostMessage(MacroMsgWindow,MM_STARTRECORD,0,0)
    else PostMessage(MacroMsgWindow,MM_STOPRECORD,0,0);
  end;
  Result:=1;  {Eat the button, even if its a button down message}
end;

procedure HandleOtherKeys;
begin
  MH:=GetCurrentMacroContext(MKey);
  If (MH<>nil) then begin
      if (MacroState=MSNone) then begin
        MacroToPlay:=GetMacroByKey(MH,MKey);
        if (MacroToPlay<>nil) then begin
          If (ButtonIsUp) then begin
            MH^.CurrentMacroRec:=MacroToPlay;
            PostMessage(MacroMsgWindow,MM_STARTPLAYBACK,0,Longint(MH));
          end;
          Result:=1;  {Eat the button, even if its a button down message}
        end;
      end;
  end;
end;

begin
  if (code<0) then Result:=CallNextHookEx(MouseHookHandle,code,wParam, lParam)
  else begin
    Result:=0;   {By default, pass the message on to Windows}
    Msg:=NormalizeMouseMsg(Wparam); {Convert non-client messages to the WM_MOUSE counterparts}

    MKey:=MessageToKey[Msg].VKey;
    ButtonIsUp:=MessageToKey[Msg].ButtonUp;
    GetShiftStates(MKey);

       {A question may be asked as to why I did not use a case statement
        here.  It would seem that

        Case MKey of
        VK_SHIFT..VK_MENU: ...
        MK_TOGGLERECORD: ...
        else
        end;

        would work nicely.  The problem is that the MK_XXXX constants are typed
        constants -- they can change at runtime.  This allows the user the ability
        to change them to different key combinations if desired.  The compiler
        evaluates the values on the left in a case statement at compile time
        so that a jump table can be constructed.  This is impossible with the MK_XXXX
        constants, hence the use of nested if's.}

      If ((CancelChecking) and (MKey=MK_CANCEL)) then HandleCancel
      else If (MKey = MK_PAUSE) then HandlePause
      else If (MKey = MK_TOGGLERECORD) then HandleRecord
      else HandleOtherKeys;
      If (Result=0) then Result:=CallNextHookEx(MouseHookHandle,code,wParam, lParam);
  end;
end;

{This routine is installed as a WH_JOURNALRECORD callback function.  It's usable as
a callback function through the use of the "export" keyword following it's declaration.
It handles the recording of key/mouse messages to the scrap macro.  If the
WindowRelativeMouse switch is set, this routine also adjusts mouse messages
it receives so that their coordinates are relative to the active Window.  When
the messages are then played back, it is the responsibility of the WH_JOURNALPLAYBACK
function to adjust them to take the current window's coordinates into consideration.
This allows the window a macro is recorded in to be moved without affecting the
accuracy of the macro.  Since mouse messages are always received with screen-relative
coordinates, a measure of this type is necessary.

It's interesting to note here that the Windows SDK documentation (and, thus, the Borland
online help for BP) for this function incorrectly states that the pointer received in
lparam points to a TMSG structure.  It points to a TEventMsg structure, just like
the WH_JOURNALPLAYBACK callback function.}

function RecordHookProc(code : integer; wparam : Word; lparam : Longint) : Longint; export;
var
  Msg : pTEventMsg;
  TempTPoint : TPoint;
  CurrentRect : TRect;
  NMessage : Word;
  VKey : Word;
begin
  try
    Msg:=pTEventMsg(lparam);
      with ScrapMacroRec^ do begin
        with MHDR do begin
          With Msg^ do begin
            NMessage:=NormalizeMouseMsg(Message);
            if ((MacroEnd<pred(MaxMacroLen)) and ((NMessage<LoMsgBound) or (NMessage>HiMsgBound))) then begin
              VKey:=lo(paraml); {Used "VKey" here because MKey is a field of ScrapMacroRec^.MHDR}
              GetShiftStates(VKey);
               if (VKey<>MK_TOGGLERECORD) and (VKey<>MK_PAUSE) then begin
                  Inc(MacroEnd);
                  Move(Msg^,Msgs[MacroEnd],sizeof(Msgs[MacroEnd]));
                  If ((NMessage >= WM_MOUSEFIRST) and (NMessage <= WM_MOUSELAST)) then begin
                    case MouseBase of
                      MBWindow : begin
                        GetWindowRect(GetActiveWindow,CurrentRect);
                        With Msgs[MacroEnd], CurrentRect do begin
                          paraml:=Msg^.paraml-left;
                          paramh:=Msg^.paramh-top;
                        end;
                      end;
                      MBPointer : begin
                        With Msgs[MacroEnd], InitialMousePosition do begin
                          paraml:=Msg^.paraml-x;
                          paramh:=Msg^.paramh-y;
                        end;
                      end;
                    end;
                  end;
               end;
            end;
          end;
        end;
      end;
    Result:=0;
  except
  On Exception Do begin
    DisplayMessage(MGUnknownError);
    Result:=MRSevereError;
  end;
  end;
end;

{This routine is installed as a WH_JOURNALPLAYBACK callback function.  It handles
the playback of key/mouse messages from a recorded macro.  If the WindowRelativeMouse
switch is set, this routine also adjusts mouse messages it playsback so that their
coordinates are relative to the active Window.  This allows the Window a macro is
recorded in to be moved without affecting the accuracy of the macro.  Since mouse
messages are always received with screen-relative coordinates, a measure of this type is
necessary.

It's interesting to note the gyrations that must be gone through to properly return
journal messages to Windows.  Windows uses basically two codes to control this
callback function:  HC_SKIP and HC_GETNEXT.  HC_SKIP tells the routine to prepare
to return a message on its next call.  HC_GETNEXT tells the routine to return a
message to Windows by copying it to the TEventMsg struct pointed to by lparam.
Knowing only this, one would thing that an HC_SKIP receipt is really a non-event,
since the index of the array containing the messages to playback can be incremented
when HC_GETNEXT is seen, but this is incorrect.  Since Windows often requests
the _same_ message more than once, an HC_SKIP receipt must be the only reason the
array index is incremented.

This is particularly true when the FastPlayback switch is set to false.  Windows
requires that the difference in time between the current message and the previous
one be returned in order to simulate playing back a macro at the same speed it was
recorded.  Once that's happended, Windows _again_ requests the same message,
only this time, the routine must _not_ return the difference between the two times,
but must instead return zero.  Returning a non-zero value, causes Windows to wait
the specified amount of time, then re-request the message.  This could cause
an infinite loop if the routine never returns zero, making the machine appear
to be hung.  The PlayingBackSlow Boolean switch is used to implement this facility.

}

function PlaybackHookProc(code : integer; wparam : Word; lparam : Longint) : Longint; export;
var
  Msg : pTEventMsg;
  TempRect : TRect;
  NMessage : Word;

begin
  try
  Result:=0;
  Msg:=pTEventMsg(lparam);
        with CurrentMacroRec^ do begin
          with MHDR do begin
            if (Position<=MacroEnd) then begin
               Case Code of
                 HC_GETNEXT : begin
                           Move(Msgs[Position],Msg^,sizeof(Msg^));
                           With Msg^ do begin
                             NMessage:=NormalizeMouseMsg(Message);
                             If ((NMessage >= WM_MOUSEFIRST) and (NMessage <= WM_MOUSELAST)) then begin
                               case MouseBase of
                                 MBWindow : begin
                                   GetWindowRect(GetActiveWindow,TempRect);
                                   With TempRect do begin
                                     paraml:=left+paraml;
                                     paramh:=top+paramh;
                                   end;
                                 end;
                                 MBPointer : begin
                                   With InitialMousePosition do begin
                                     paraml:=x+paraml;
                                     paramh:=y+paramh;
                                   end;
                                 end;
                               end;
                             end;
                           end;
                           If not(FastPlayback) then begin
                             If (not(PlayingBackSlow)) then begin
                               if (Position<>0) then Result:=(Msgs[Position].time-Msgs[pred(Position)].time)
                               else Result:=0;
                               PlayingBackSlow:=true;
                             end else begin
                               PlayingBackSlow:=false;
                               Result:=0;
                             end;
                           end else Result:=0;
                 end;
                 HC_SKIP : begin
                             If (CancelChecking) and (BitSet(GetAsyncKeyState(MK_CANCEL),1)) then
                                PostMessage(MacroMsgWindow, MM_STOPPLAYBACK,0,0);
                             Inc(Position);
                           end;
                 HC_SYSMODALON : SendMessage(MacroMsgWindow,MM_PAUSEPLAYBACK,0,0);
                 HC_SYSMODALOFF : SendMessage(MacroMsgWindow,MM_RESUMEPLAYBACK,0,0);
               end;
            end else begin
              If (Code=HC_SKIP) then SendMessage(MacroMsgWindow,MM_STOPPLAYBACK,0,0);
            end;
          end;
        end;
  except
  On Exception Do begin
    DisplayMessage(MGUnknownError);
    Result:=MRSevereError;
  end;
  end;
end;

{This routine sets the playback hook}

procedure HookPlayback;
begin
  GetAsyncKeyState(MK_CANCEL); {Init GetAsyncKeyState's "previously pressed" toggle}
  PlaybackHookHandle:=SetWindowsHookEx(WH_JOURNALPLAYBACK,PlaybackHookProc,HInstance,0);
end;

{This routine unhooks the playback routine}

procedure UnhookPlayback;
begin
   UnHookWindowsHookEx(PlaybackHookHandle);
end;

{Plays back the macro passed into it.}

procedure PlaybackMacro(const MacroToPlay : pTMacroRec); export;
begin
  try
  If (MacroState=MSNone) then begin
    If (MacroToPlay<>nil) then begin
      MacroState:=MSPlayback;
      With MacroToPlay^.MHDR do
        If (@CodeHook=nil) then begin
          Position:=0;
          SaveToggles:=GetToggleStates;
          SetToggleStates(MacroToPlay^.MHDR.Toggles);
          If (MouseBase=MBPointer) then GetCursorPos(InitialMousePosition);
          BroadcastMacroMsg(MM_PLAYBACKON,UMSNone,0);
          BroadcastMacroMsg(MM_PLAYBACKCHANGE,PSOn,0);
          CurrentMacroRec:=MacroToPlay;
          HookPlayback;
        end else begin
          CodeHook;     {Call the routine stored in the CodeHook field}
          MacroState:=MSNone;
          SetKeyboardState(SaveKeyboardState);
        end;
    end;
  end else DisplayMessage(MGCantDoNow);
  except
  On Exception Do begin
    DisplayMessage(MGUnknownError);
  end;
  end;
end;

procedure TypeKeys(Keys : PChar); export;
var
  KeysLen : Word;
  KeysLongint : Longint;
  CurrentWindow : HWND;
  KCnt : Word;
  ScanCode : Word;
  VKey : Word;
begin
  KeysLongint:=0;
  CurrentWindow:=GetFocus;
  KeysLen:=pred(StrLen(Keys));
  For KCnt:=0 to KeysLen do begin
      ScanCode:=LoWord(OemKeyScan(ord(Keys[Kcnt])));
      If (ScanCode>0) then begin
        VKey :=lo(vkKeyScan(ord(Keys[Kcnt])));
        If (VKey>0) then begin
          LBytes(KeysLongint)[2]:=ScanCode;
          PostMessage(CurrentWindow,WM_CHAR,VKey,KeysLongint);
        end;
      end;
  end;
end;

{Copies the parameters stored in a macro context to their global counterparts.
This allows macros to be recorded and played back without the related routines
having to be aware of the "current" macro context.  I've considered putting these
global variables into one global "context" -- this would allow me to easily move one to
the other.}

procedure CopyParmsToGlobal(const MH : pTMacroContext);
begin
  CurrentMacroRec:=MH^.CurrentMacroRec;
  CallingWindowHook:=MH^.CallingWindowHook;
  MaxMacroNum:=MH^.MaxMacroNum;
  FastPlayBack:=MH^.FastPlayback;
  TaskScope:=MH^.TaskScope;
  LoMsgBound:=MH^.LoMsgBound;
  HiMsgBound:=MH^.HiMsgBound;
  MouseBase:=MH^.MouseBase;
  CancelChecking:=MH^.CancelChecking;
end;

{Turns recording on}

procedure RecordingOn; export;
begin
  try
  If (MacroState = MSNone) then begin
             MacroState:=MSRecord;
             ScrapMacroRec^.MHDR.Toggles:=GetToggleStates;
             ScrapMacroRec^.MHDR.MacroEnd:=-1;
             ScrapMacroRec^.MHDR.MouseBase:=MouseBase;
             If (MouseBase=MBPointer) then GetCursorPos(InitialMousePosition);
             BroadcastMacroMsg(MM_RECORDON,UMSNone,0);
             BroadcastMacroMsg(MM_RECORDCHANGE,RSOn,0);
             RecordHookHandle:=SetWindowsHookEx(WH_JOURNALRECORD,RecordHookProc,HInstance,0);
  end else DisplayMessage(MGCantRecord);
  except
  On Exception Do begin
    DisplayMessage(MGUnknownError);
  end;
  end;
end;

{Turns recording off}

procedure RecordingOff; export;
begin
  try
    If ((MacroState=MSRecord) or (MacroState=MSPauseRecord)) then begin
      UnhookWindowsHookEx(RecordHookHandle);
      MacroState:=MSNone;
      BroadcastMacroMsg(MM_RECORDOFF,UMSNone,0);
      BroadcastMacroMsg(MM_RECORDCHANGE,RSOff,0);
    end else DisplayMessage(MGNotRecording);
  except
  On Exception Do begin
    DisplayMessage(MGUnknownError);
  end;
  end;
end;

{Toggles macro recording.  Note the use of an array of procedures to effect
a simple jump table.  Also note the typecasting of MacroState in order to handle
it like a Boolean toggle.  This is possible because MSNone and MSRecord occupy the
zero and one slots, respectively, in MacroState.  Toggling the variable as a
Boolean either sets it to false (MSNone) or true (MSRecording), which is what
we want for use with our jump table.}

procedure ToggleRecording; export;
begin
  try
    case MacroState of
    MSNone : RecordingOn;
    MSRecord, MSPauseRecord : RecordingOff;
    else DisplayMessage(MGNotRecording);
    end;
  except
    On Exception do DisplayMessage(MGUnknownError);
  end;
end;

procedure PauseRecording; export;
begin
  If (MacroState = MSRecord) then begin
    MacroState:=MSPauseRecord;
    BroadcastMacroMsg(MM_RECORDOFF,UMSNone,0);
    BroadcastMacroMsg(MM_RECORDCHANGE,RSPaused,0);
    UnhookWindowsHookEx(RecordHookHandle);
  end else DisplayMessage(MGNotRecording);
end;

procedure ResumeRecording; export;
begin
  If (MacroState = MSPauseRecord) then begin
    MacroState:=MSRecord;
    BroadcastMacroMsg(MM_RECORDON,UMSNone,0);
    BroadcastMacroMsg(MM_RECORDCHANGE,RSResumed,0);
    RecordHookHandle:=SetWindowsHookEx(WH_JOURNALRECORD,RecordHookProc,HInstance,0);
  end else DisplayMessage(MGNotRecording);
end;

{Toggle a temporary pause when recording macros.}

procedure TogglePause; export;
begin
  case MacroState of
    MSRecord : PauseRecording;
    MSPauseRecord : ResumeRecording;
    else DisplayMessage(MGNotRecording);
  end;
end;

Procedure ProcessMessages;
Var
  Msg : Tmsg;
begin
  If PeekMessage(Msg,0, WM_KEYFIRST, WM_KEYLAST, PM_REMOVE) then begin
     TranslateMessage(Msg);
     DispatchMessage(Msg);
  end;
end;


{This routine acts as the central command processor for the macro engine.  It dispatches
messages it receives by calling the appropriate routines.  It also determines if
shift keys are still down when it receives a message and waits for them to be released
before proceeding.}

function MacWinProc(Window: HWnd; Message, WParam: Word; LParam: Longint): Longint;
var
  ShiftStates : Word;
  MH : pTMacroContext;
  Msg : TMsg;
  RecordCanceled : Boolean;

Procedure RemoveMsg(Macro : pTmacroRec; Msg, Key : Word);
var
  Ind : Integer;
  MCnt : Word;
begin
  Ind:=-1;
  With Macro^ do begin
    For MCnt:=MHDR.MacroEnd downto 0 do
      If Msgs[mcnt].Message=Msg then begin
        If ((Key<>0) and (lo(Msgs[mcnt].paraml)=Key)) or (Key=0) then begin
          Ind:=Mcnt;
          break;
        end;
      end;
    If (Ind<>-1) then begin
      If (Ind<>MHDR.MacroEnd) then
        For Mcnt:=Ind to pred(MHDR.MacroEnd) do
          Msgs[Mcnt]:=Msgs[Succ(Mcnt)];
      Dec(MHDR.MacroEnd);
    end;
  end;
end;

begin
  Result := 0;
  MH := pTMacroContext(lparam);
      case Message of
        MM_STARTRECORD : begin
                           GetKeyBoardState(SaveKeyboardState);
                           ClearShiftStates;
                           RecordingOn;
                         end;
        MM_STOPRECORD : begin
                          RecordCanceled := WordBool(WParam);
                          If RecordCanceled then ScrapMacroRec^.MHDR.MacroEnd:=-1;
                          RecordingOff;
                          SetKeyBoardState(SaveKeyboardState);
                        end;
        MM_STARTPAUSE : begin
                          ClearShiftStates;
                          PauseRecording;
                          If bytebool(hi(MK_PAUSE) and AltOn) then begin
                            RemoveMsg(ScrapMacroRec,WM_SYSKEYDOWN,0);
                            RemoveMsg(ScrapMacroRec,WM_KEYUP,VK_MENU);
                          end;
                        end;
        MM_STOPPAUSE : begin
                         ResumeRecording;
                       end;
        MM_STARTPLAYBACK : begin
                             MH^.MouseBase:=MH^.CurrentMacroRec^.MHDR.MouseBase;
                             CopyParmsToGlobal(MH);
                             GetKeyBoardState(SaveKeyboardState);
                             ClearShiftStates;
                             PlaybackMacro(CurrentMacroRec);
                           end;
        MM_STOPPLAYBACK : If (MacroState=MSPlayback) then begin
                            MacroState:=MSNone;
                            UnhookPlayback;
                            BroadcastMacroMsg(MM_PLAYBACKOFF,UMSNone,0);
                            BroadcastMacroMsg(MM_PLAYBACKCHANGE,PSOff,0);
                            SetToggleStates(SaveToggles);
                            SetKeyboardState(SaveKeyboardState);
                          end;
        MM_PAUSEPLAYBACK : If (MacroState=MSPlayback) then begin
                            UnhookPlayback;
                            BroadcastMacroMsg(MM_PLAYBACKPAUSED,UMSNone,0);
                            BroadcastMacroMsg(MM_PLAYBACKCHANGE,PSPaused,0);
                            SetToggleStates(SaveToggles);
                          end;
        MM_RESUMEPLAYBACK : If (MacroState=MSPlayback) then begin
                            HookPlayback;
                            SetToggleStates(ScrapMacroRec^.MHDR.Toggles);
                            BroadcastMacroMsg(MM_PLAYBACKRESUMED,UMSNone,0);
                            BroadcastMacroMsg(MM_PLAYBACKCHANGE,PSResumed,0);
                          end;
      else
        Result := DefWindowProc(Window, Message, WParam, LParam);
      end;
end;

{Returns a pointer to the scrap macro.}

function GetScrapMacro: pTMacroRec; export;
begin
  Result:=ScrapMacroRec;
end;

{Plays the scrap macro back.}

procedure PlaybackScrap; export;
begin
  try
  PlaybackMacro(ScrapMacroRec);
  except
  On Exception Do begin
    DisplayMessage(MGUnknownError);
  end;
  end;
end;

{Builds the macro message handling Window.}

function CreateMacroWindow : HWND;
begin
  Result := CreateWindow(MacClass.lpszClassName,EmptyPChar,0,
  cw_usedefault,cw_usedefault,cw_usedefault,cw_usedefault,
  HWND_DESKTOP,0,
  HInstance,nil);
end;

{Gets the value of MacroState, used for exporting to the outside world.}

function GetMacroState : TMacroState; export;
begin
  Result:=MacroState;
end;

{Forward declaration for our ShutDown procedure}

procedure ShutdownWinMacro; forward;


{This procedure inializes the library's various global data structures.
Our window class is registered and the keyboard and mouse hooks are set.

This routine is only called if it has not yet been called.  The LibraryInitialized
variable is used to determine whether it has yet been successfully called.}

procedure InitializeWinMacro;
begin
  try
  FillChar(MacroContexts,sizeof(MacroContexts),0);

  KeyboardHookHandle:=0;
  MouseHookHandle:=0;
  MacroMsgWindow:=0;
  ScrapMacroRec:=nil;

  MacClass.hInstance := HInstance;
  if not(RegisterClass(MacClass)) then begin
     DisplayMessage(MGRegisterClass);
     ShutDownWinMacro;
     Halt(1);
  end;

  KeyboardHookHandle := SetWindowsHookEx(WH_KEYBOARD,KeyboardHookProc,HInstance,0);
  if (KeyboardHookHandle=0) then begin
      DisplayMessage(MGHookKeyboard);
      ShutDownWinMacro;
      Halt(1);
  end;

  MouseHookHandle := SetWindowsHookEx(WH_MOUSE,MouseHookProc,HInstance,0);
  if (MouseHookHandle=0) then begin
      DisplayMessage(MGHookMouse);
      ShutDownWinMacro;
      Halt(1);
  end;

  MacroMsgWindow:=CreateMacroWindow;
  If (MacroMsgWindow=0) then begin
    DisplayMessage(MGCreateMsgWindow);
    ShutDownWinMacro;
    Halt(1);
  end;

  MaxMacroNum:=0;
  MaxMacroLen:=0;
  PlayingBackSlow:=false;
  CallingWindowHook:=0;
  Position:=0;

  LibraryInitialized:=true;
  except
  On Exception Do begin
    DisplayMessage(MGUnknownError);
    ShutDownWinMacro;
    Halt(1);
  end;
  end;
end;

{
function InstallHook(const _CallingWindowHook : HWND; Const _MaxMacroNum, _MaxMacroLen : Word;
                     const _FastPlayback, SystemWide : Boolean;
                     const _RecordMouse : TMouseEvent;
                     const _MouseBase : TMouseBase;
                     const _CancelChecking : Boolean) : pTMacroContext; export;

  CallingWindow is the window handle of the calling window.  It is needed later
  by the engine to send messages when the state of the engine changes.  For example,
  when recording is toggled, the engine sends a message to the calling window,
  advising it that a state change has occurred relating to recording and what that
  change is.

  MaxMacNum and MaxMacLen specify the maximum number of macros allowed and the
  maximum length of each one, respectively.  Note that MaxMacLen DOES NOT establish
  the number of _keys_ that may be recorded to a macro, but rather the number of
  keyboard (and mouse) _messages_ that may be recorded.  A single keypress will
  always be equivalent to at least two messages:  one for the key down message and
  one for the key up message.  Shift key combinations require separate messages
  to be recorded for each press and release of the shift key.  Also remember that
  mouse movements generate lots of messages very quickly.  See the RecordMouse
  parameter below for a way to limit which mouse messages are recorded.

  Fastplay determines how macros are played back, either as quickly as possible
  (the default) or at the same speed they were recorded.

  SystemWide determines whether a recorded macro will have system wide scope
  or will only playback in the context of the task that called InstallHook.

  RecordMouse establishes what types of mouse events will be recorded.  The macro
  engine can be set up to record all mouse events, only clicks and drags or no
  mouse events.

  MouseBase estalishes whether the pointer coordinates in mouse messages
  are seen as relative to the entire screen, to the current window or to the
  initial position of the mouse pointer.  The macro engine plays back mouse
  messages in screen-relative macros using the same pointer coordinates with
  which they were recorded.  This means that mouse messages played back to a
  window that has been moved will probably not playback correctly, since the
  windows controls move with it and are no longer in the positions they occupied
  when the macro was recorded.  The answer to this problem is the window-relative
  macro.  It adjusts mouse messages to take the position of the current window
  into account as they are recorded.  When they are played back, the engine again
  causes the messages to be based on the coordinates of the upper left corner of
  the current window.

  Although the ability to base mouse message recording and playback on either the
  screen or current window covers most users' needs, there is one situation where
  they do not.  This situation is most often encountered in paint or draw
  programs.  If a user wanted to record a macro that drew, say, a box in the
  current draw context, neither a screen-relative macro, nor a window-relative
  macro would do the trick.  A screen-relative macro would cause the macro to
  attempt to draw a box at exactly the same spot on the screen each time it was
  played back.  A window-relative macro would draw the box in exactly the same
  spot in the current window each time the macro was played back.  The answer to
  this problem is the pointer-relative macro.  It adjusts all mouse messages
  recorded or played back to be based on the current position of the mouse pointer.
  All the user in the aforementioned example would have to do to draw several boxes
  using his newly created macro would be to move the pointer, and simply replay
  the macro.

  The MouseBase flag is maintained at three levels for maximum flexibility.  It is
  stored in each macro record so that each macro in a macro file can have its own
  individual base.  It is stored in each macro context so that newly recorded macros
  can have a default base as specified by the developer.  It's stored at the global
  level so that the macro engine need not track a "current macro context" or the
  like when recording or playing back macros.

  CancelChecking determines whether the engine watches for the pressing of a key that
  cancels macro playback.  Because of the way Windows' journaling system works and, by
  extension, the way the cancel key must be detected, it must be unshifted: i.e.,
  require no shift keys to be pressed simultaneously with it.

returns

  InstallHook returns a handle to a macro "context" that must be used in subsequent
  calls to the engine.  It's important to store this value in a variable since
  you really can't do anything without it, including shut down the engine itself.

}
function InstallHook(const _CallingWindowHook : HWND; Const _MaxMacroNum, _MaxMacroLen : Word;
                     const _FastPlayback, SystemWide : Boolean;
                     const _RecordMouse : TRecordMouse;
                     const _MouseBase : TMouseBase;
                     const _CancelChecking : Boolean) : pTMacroContext; export;

var
  MH : pTMacroContext;
begin
  Result:=nil;
  try
  If (MacroState=MSNone) then begin
    If (not(LibraryInitialized)) then InitializeWinMacro;
    GetMemCheck(MH, sizeof(TMacroContext));
    If (MH<>nil) then begin
      MH^.TaskHandle:=GetCurrentTask;
      If not(AddMacroContextEntry(MH)) then
        FreeMemCheck(MH, sizeof(TMacroContext))
      else begin
          If not GetMemCheck(MH^.Macros,_MaxMacroNum*sizeof(pTMacroRec)) then begin
            DisplayMessage(MGAllocateContext);
            Exit;
          end;
          FillChar(MH^.Macros^,_MaxMacroNum*sizeof(pTMacroRec),0);

          If (_MaxMacroLen>MaxMacroLen) or (ScrapMacroRec=nil) then begin
             MaxMacroLen:=_MaxMacroLen;
             If (ScrapMacroRec<>nil) then begin
               {Save off current hot key in case it has been changed}
               MK_PLAYBACKSCRAP:=ScrapMacroRec^.MHDR.MKey;
               DeallocateMacroPrim(ScrapMacroRec);
             end;
             If not GetMemCheck(ScrapMacroRec,sizeof(TMHDR)+(MaxMacroLen*sizeof(TEventMsg))) then begin
               DisplayMessage(MGAllocateScrap);
               Exit;
             end;
             If (ScrapMacroRec=nil) then {Error box already displayted} Exit;
             With ScrapMacroRec^.MHDR do begin
               Name:=nil;
               MKey:=MK_PLAYBACKSCRAP;
               Toggles:=0;
               SetBit(Toggles,ScrapMacroStatusBit);
               @CodeHook:=nil;
               MacroEnd:=-1;
             end;
          end;

          MH^.MaxMacroLen:=MaxMacroLen;
         With MH^ do begin
           FastPlayback:=_FastPlayback;
           If (SystemWide) then TaskScope:=0
           else TaskScope:=GetWindowTask(_CallingWindowHook);
           MaxMacroNum:=_MaxMacroNum;
           CallingWindowHook:=_CallingWindowHook;
           FastPlayback:=_FastPlayback;
           Case _RecordMouse of
           RMIgnore : begin
                           LoMsgBound:=WM_MOUSEFIRST;
                           HiMsgBound:=WM_MOUSELAST;
                         end;
           RMClicksAndDrags : begin
                              LoMsgBound:=WM_MOUSEMOVE;
                              HiMsgBound:=WM_MOUSEMOVE;
                            end
           else begin
                    LoMsgBound:=WM_NULL;
                    HiMsgBound:=WM_NULL;
                  end;
           end;

           MouseBase:=_MouseBase;
           CancelChecking:=_CancelChecking;
         end;
         CopyParmsToGlobal(MH);
      end;
    end;
    Result:=MH;
  end else DisplayMessage(MGCantDoNow);
  except
  On Exception Do begin
    DisplayMessage(MGUnknownError);
    Result:=nil;
  end;
  end;
end;

{This routine reverses the effects of the InitializeWinMacro initialization routine.}

procedure ShutdownWinMacro;
var
  TempInfo : TWNDClass;
  Ccnt : Byte;
begin
  try

  If (KeyboardHookHandle<>0) and (not UnhookWindowsHookEx(KeyboardHookHandle)) then
    DisplayMessage(MGUnhookKeyboard);

  If (MouseHookHandle<>0) and (not UnhookWindowsHookEx(MouseHookHandle)) then
    DisplayMessage(MGUnhookMouse);

  If (ScrapMacroRec<>nil) then DeallocateMacroPrim(ScrapMacroRec);

  If (MacroMsgWindow<>0) and (not DestroyWindow(MacroMsgWindow)) then
    DisplayMessage(MGDestroyMsgWindow);

  If (GetClassInfo(hInstance,MacClass.lpszClassName,TempInfo)) and (not UnregisterClass(MacClass.lpszClassName,hInstance)) then
    DisplayMessage(MGUnregisterClass);

  LibraryInitialized:=false;
  except
  On Exception Do begin
    DisplayMessage(MGUnknownError);
  end;
  end;
end;

function CheckActiveContexts : Boolean;
var
  Ccnt : Byte;
begin
  Result:=false;
  For Ccnt:=0 to MaxMacroContext do if (MacroContexts[Ccnt]<>nil) then begin
    Result:=true;
    Break;
  end;
end;

{Deallocates a macro context}

procedure UninstallHook(var MH : pTMacroContext); export;
var
  Mcnt, MHIndex : Byte;
begin
  try
  If (MacroState=MSNone) then begin
    If CheckMacroHandle(MH) then With MH^ do begin
      If (Macros=nil) then DisplayMessage(MGNoMacrosInstalled)
      else begin
        DeallocateMacros(MH);
        FreeMemCheck(Macros,MaxMacroNum*sizeof(pTMacroRec));
      end;

      MHIndex:=GetMacroContextIndex(MH);
      If Not FreeMemCheck(MH,sizeof(TMacroContext)) then
        DisplayMessage(MGDeallocateContext);
      If (MHIndex<>-1) then MacroContexts[MHIndex]:=nil;
      If (not(CheckActiveContexts)) then ShutdownWinMacro;
    end;
  end else DisplayMessage(MGCantDoNow);
  except
  On Exception Do begin
    DisplayMessage(MGUnknownError);
  end;
  end;
end;

{Returns the version of the library.  Can be used by applications to ensure
they do not have an application/DLL version mismatch.}

function GetWinmacroVersion : Word; export;
begin
 Result:=Version;
end;


exports
  LoadMsg,
  GetKeyCtrl,
  SetKeyCtrl,
  GetKeyAlt,
  SetKeyAlt,
  GetKeyShift,
  SetKeyShift,
  GetShiftStates,
  GetPossibleMKey,
  GetPossibleMKeyIndex,
  GetMaxPossibleMKeys,
  GetMacroByKey,
  GetMacroByIndex,
  GetMacroKey,
  SetMacroKey,
  GetMacroName,
  SetMacroName,
  GetMacroCodeHook,
  SetMacroCodeHook,
  GetToggleRecordKey,
  SetToggleRecordKey,
  GetCancelKey,
  SetCancelKey,
  GetPauseKey,
  SetPauseKey,
  GetKeyName,
  WriteMacros,
  ReadMacros,
  ToggleRecording,
  RecordingOn,
  RecordingOff,
  GetMacroState,
  InstallMacro,
  PlaybackMacro,
  TypeKeys,
  GetScrapMacro,
  PlaybackScrap,
  InstallHook,
  UninstallHook,
  GetWinmacroVersion;

begin
end.

