{$I WINMACRO.INC}
{
WMacro unit for AppBuilder.  Provides keyboard and macro playback facilities
to Windows-based apps via the TMacroSet component.

Written by Ken Henderson.

Copyright (c) 1994 Ken Henderson.  All Rights Reserved.

Last modification: 11-29-94

Components

This unit defines the TMacroSet, TMacroKeyEdit and TMacroKeyComboBox components.
TMacroSets are sets of recorded macros.  The number of macros and the maximum size
of each is set via properties.  The unit requires WINMACRO.DLL, which contains the
actual macro engine, and WMTypes.PAS and WMImport.PAS which function as the
interface to the engine.

TMacroKeyEdit is a descendant of the TEdit class.  It allows the easy entry of
macro hotkeys -- the names of which are stored in the component's text field as
they are typed.  The component's MKey field is set appropriately based on what keys
the user presses.

TMacroKeyComboBox performs the same function as the TMacroKeyEdit component in that
it allows the easy entry of macro hotkeys.  It also allows them to be selected from
a drop-down list of all available keys.

This unit also registers custom property editors for setting various special
keys.

Order of events

Applications interact with this component by first calling the MacrosOn routine.
An application would typically call it from within a form's OnCreate handler,
though calling it at other times works as well.

Next, a keyboard macro can be recorded by pressing the toggle record key or
turning recording on via a call to either RecordingOn or ToggleRecording.

Finally, the recorded macro (which is stored initially as the scrap macro) can
be copied to a permanent macro via a call to TMacroSet's Add method.

Code macros (macros that execute code as opposed to playback keys) can be assigned
without recording anything before calling Add, though a valid macro pointer still
needs to be supplied to the Add method.  The scrap macro is always valid for this
use even when nothing has yet been recorded to it.

MacrosOff is called to shutdown the macro engine.  An application would typically
call it from within a form's OnClose handler, though calling it at other times
works as well.

Other notes

Only one component may be installed per application.  This is primarily due to
the fact that the engine uses the current application's task handle to help it
find allocated macros.

Much of the functionality of the Windows Recorder accessory is included, but
TMacroSet's far exceed Recorder's abilities.  For one thing, TMacroSet's allow
mouse buttons (and combinations of mouse buttons and keys), not just keyboard keys
to act as hotkeys.  Also, the number and maximum length of macros is configurable
via properties, saving memory resources.  Also, macros can be "code" macros --
macros that execute procedures passed into them -- rather than just keyboard macros.
This ability, along with the ability to tie a macro to any key on the keyboard,
allows the macro engine to be used to as a command processor, essentially establishing
accelerator keys at runtime.  These "accelerator" assignments can be changed at
will while the program is running.

Since exceptions cannot be communicated from DLL's back to their callers, the
underlying DLL, Winmacro, sets up open Try...Except blocks around every entry
point into the DLL to keep exceptions from causing the DLL to abort and take
the calling application with it.  When such exceptions are seen, the DLL sets
the return value of the offending function to the most appropriate error code
value and exits the routine.  On this end -- within this component -- that return
code is translated back to the appropriate exception and re-raised, effectively causing
the exception to bridge the DLL/application barrier.  It is the calling application's
responsibility to trap these exceptions.  They are listed below.

Using the indexed properties facility, many of the mundane and tedious aspects
of the macro engine, which exist of necessity, are hidden from the developer.  An
example of this is the setting of the shift key bit flags for hotkeys (MKeys).
The high order byte of MKey word variables stores the status of the shift key flags
used in conjunction with the main key.  For example, bit 1 is set if Control is
used with the key, bit 2 if Alt is used and Shift sets bit 3.  Working directly
with the DLL would require that the programmer call the exported routines for
getting and setting these bit flags.  Using indexed properties, though, we are
able to effect a psuedo-array for each of the toggles and easily toggle them.
For example to set the Control bit on for a given MKey we:

Controls[MKey]:=true;

using the MKey as an index into the psuedo-array of Control flag switches corresponding
to all allocated MKeys.  No direct call to the DLL is then necessary.

This convention, that of using indexed properties to insulate the developer from
the raw aspects of the macro engine DLL, is used extensively in this component.

}
Unit WMacro;

interface

Uses WinTypes,
  Controls,
  Classes,
  StdCtrls,
  SysUtils,
  WinProcs,
  Messages,
  Forms,
  DsgnIntf,
  TypInfo,
  WMTypes,
  WMImport;

{Registration procedure, required of all components}

Procedure Register;


const

  {Property defaults}
  {$I WMDFAULT.INC}

  DefaultCallingWindow=0;
  DefaultMaxMacros=10;
  DefaultMaxMacroLen=30;
  DefaultFastPlayback=true;
  DefaultSystemWide=true;
  DefaultRecordMouse=RMIgnore;
  DefaultMouseBase=MBWindow;
  DefaultCancelChecking=false;
  DefaultCancelKey=DEFAULTMK_CANCEL;
  DefaultPauseKey=DEFAULTMK_PAUSE;
  DefaultToggleRecordKey=DEFAULTMK_TOGGLERECORD;
  DefaultMKey=0;

Type

  {Exception classes}

  EHookError = class(Exception);
  EMacroNotFound = class(Exception);
  EInstallMacro = class(Exception);
  EWriteMacros = class(Exception);
  EReadMacros = class(Exception);
  EShiftKeysNotAllowed = class(Exception);

  {Specialized notify events}

  TNotifyStateEvent = procedure (Sender: TObject; State : TUserMacroState) of object;

  TNotifyAssignEvent = procedure (Sender: TObject; Macro : pTMacroRec) of object;

  {TMKeyProperty property editor class

  This class allows us to edit MKey values using a list of the key names of all
  possible keys.}

  TMKeyProperty = class(TPropertyEditor)
  public
    {Tells the Object Inspector that this property provides a list of values}
    function GetAttributes: TPropertyAttributes; override;
    {Returns a single MKey as a string}
    function GetValue: string; override;
    {Pops all possible MKeys on to a list internal to the Object Inspector using
    the provided TGetStrProc routine}
    procedure GetValues(Proc: TGetStrProc); override;
    {Sets a single MKey value by converting its string representation back to a
    valid MKey}
    procedure SetValue(const Value: string); override;
    {Returns the textual name of a given MKey as a string}
    function GetKeyName(const MKey : Word) : string;
  end;

  {TMUnshiftedKeyProperty property editor class

  This class allows us to edit MKey values using a list of the key names of all
  possible keys.}

  TMUnshiftedKeyProperty = class(TPropertyEditor)
  public
    {Tells the Object Inspector that this property provides a list of values}
    function GetAttributes: TPropertyAttributes; override;
    {Returns a single MKey as a string}
    function GetValue: string; override;
    {Pops all possible MKeys on to a list internal to the Object Inspector using
    the provided TGetStrProc routine}
    procedure GetValues(Proc: TGetStrProc); override;
    {Sets a single MKey value by converting its string representation back to a
    valid MKey}
    procedure SetValue(const Value: string); override;
    {Returns the textual name of a given MKey as a string}
    function GetKeyName(const MKey : Word) : string;
  end;

  {TMacroSet component class

  Concepts

  This class defines a component used to manage a list of keyboard/mouse/code
  macros.  Two types of macros can be recorded, those that playback keyboard or
  mouse messages ("keyboard" macros) and those that execute callback functions
  ("code" macros).  Keys, mouse buttons or combinations of the two can be used
  to trigger macro playback.  Additionally, macros can be played by calling the
  appropriate class methods.  Macro hotkeys are stored as "MKeys" -- word variables
  that contain a Windows virtual keycode in the low order byte and the status of
  the "transient" shift keys (Control, Alt and Shift) in the high order byte.
  The status byte is broken down as follows:

  Switch	Bit
  ------       ---
  Control	  1
  Alt		  2
  Shift	  3

  bits 4-6 are used to store the status of the toggled shift keys (Capslock, Numlock
  and ScrollLock) when appropriate, but these never make up part of a macro hot
  key.  They are saved when a macro is recorded, and restored when it plays back
  to ensure that the messages recorded in the macro playback accurately.  The class
  defines a rich suite of methods to manipulate these bits via indexed properties.
  See the discussion on indexed properties below for more information.

  The macros themselves are divided into two parts:  the header and the message
  sections.  The header lists all fields in the macro except the recorded
  messages themselves.  This design allows us to alter the component fields of
  a macro without invalidating the routines elsewhere in the macro engine that
  calculate the amount of memory needed to store a macro.  Since the engine only
  allocates enough memory for a new macro to store the number of messages
  recorded in the source macro (usually the scrap macro), this calculation must
  be exact.

  Properties

  FMacroContext is used by the component to store the macro context
  handle returned by the macro engine.  It is used in all subsequent calls
  to the engine, with few exceptions.  Routines that merely perform bit-twiddling
  operations on MKey's do not require the services of the engine itself, so they
  do not require a context handle.  Routines that are not context specific,
  like turning on macro recording, do not require one either.

  FCallingWindow stores the handle of the message processing Window that the
  component establishes.  It creates a Window for processing messages received
  from the macro engine.  These messages are, in turn, translated into TNotifyEvents
  so that any code the user has hooked to an event can be executed.  The FCallingWindow
  HWND is passed into the macro engine via the call to InstallHook in the class
  constructor.

  Events

  Nine TNotifyEvents are established by the class.  FOnRecord is triggered
  when recording begins, FOnRecordStop when it ends.  FOnPlay is triggered
  when playback begins, FOnPlayEnd when it ends.  FOnPaused when macro
  playback is paused (caused only when a system modal dialog box displays),
  FOnResumed when playback is resumed following a pause.  FOnPlayed is
  triggered whenever the status of macro playback changes (it begins or
  ends) and passes the new status along with the notify event.  Likewise,
  FOnRecorded is triggered when the status of macro recording changes; it
  also provides the new status as part of the notify event.  FOnAssigned
  is triggered when a new macro assignment occurs.  This can be triggered
  via a call to either the component's Add or its ReadFromFile method.

  Some notify events occur in tandem.  That is, turning recording on will
  cause both FOnRecord and FOnRecorded to occur.  FOnRecorded also
  accompanies FOnRecordStop.  The somewhat obscure, inconsistent naming of
  these events came about because of length limitations within the Object
  Inspector and the desire to keep them consistent with their published
  counterparts.

  }

  TMacroSet = class(TComponent)
  private
    FMacroContext : pTMacroContext;
    FCallingWindow : HWND;
    FMaxMacros, FMaxMacroLen : word;
    FFastPlayback, FSystemWide : boolean;
    FRecordMouse : TRecordMouse;
    FMouseBase : TMouseBase;
    FCancelChecking : Boolean;
    FVersion : String;
    FOnRecord,
    FOnRecordStop,
    FOnPlay,
    FOnPlayEnd,
    FOnPaused,
    FOnResumed : TNotifyEvent;
    FOnPlayed,
    FOnRecorded : TNotifyStateEvent;
    FOnAssigned : TNotifyAssignEvent;
    procedure WndProc(var Message: TMessage);
  protected
    function GetNumberedMacro(const Index : Integer) : pTMacroRec;
    function GetNamedMacro(const Name : string) : pTMacroRec;
    function GetKeyedMacro(const MKey : Word) : pTMacroRec;
    function GetMacroName(const MKey : Word) : string;
    procedure SetMacroName(const MKey : Word; const Name : String);
    function GetMacroKey(const Index : Integer) : Word;
    procedure SetMacroKey(const Index : Integer; const MKey : Word);
    function GetKeyName(const MKey : Word) : string;
    function GetCodeHook(const MKey : Word) : CodeHookProc;
    procedure SetCodeHook(const MKey : Word; const CodeHook : CodeHookProc);
    function GetScrapMacro : pTMacroRec;
    function GetMacroState : TMacroState;
    function GetToggleRecordKey : Word;
    procedure SetToggleRecordKey(const MKey : Word);
    function GetCancelKey : Word;
    procedure SetCancelKey(const MKey : Word);
    function GetPauseKey : Word;
    procedure SetPauseKey(const MKey : Word);
    function GetPossibleKey(const Index : Word) : Word;
    function GetMaxPossibleKeys : Word;
    function GetControl(var MKey : word) : boolean;
    procedure SetControl(var MKey : Word; const Ctrl : boolean);
    function GetAlt(var MKey : word) : boolean;
    procedure SetAlt(var MKey : Word; const Alt : boolean);
    function GetShift(var MKey : word) : boolean;
    procedure SetShift(var MKey : Word; const Shift : boolean);
    function GetStateName(const Index : TUserMacroState) : String;
    function GetVersionString : string;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    {Copies a previously recorded keyboard macro to a new one or allocates a code macro.}
    procedure Add(const MKey : word;
                      const Macro : pTMacroRec;
                      const Name : String;
                      const CodeHook : CodeHookProc);

    {Deletes an allocated macro.  It does this by calling the same underlying
    routine as Add, but passes Nil as the Macro parameter}
    procedure Delete(const MKey : Word);

    procedure ToggleRecording;
    procedure SetRecordingOn;
    procedure SetRecordingOff;
    procedure PlaybackMacro(const MKey : Word);
    procedure TypeKeys(const KeyChars : PChar);
    procedure PlaybackScrap;

    {Writes all currently allocated macros to a file}
    procedure SaveToFile(const FileName : String);

    {Reads the macros contained in the specified file.  If Merge it set, it
    adds them to the macros already installed}
    procedure ReadFromFile(const FileName : String; const Merge : boolean);

    {MacrosOn must be called prior to any functions which actually use the macro
    engine.  It saves off the FMacroContext handle for later interaction with the
    engine.}
    function MacrosOn : boolean;

    {MacrosOff must be called before an application terminates in order to
    release all allocated memory and release any hooks into Windows.}
    procedure MacrosOff;

    {Macros returns a pointer the individual macro pointer (see WMTypes for more
    information) corresponding to a given MKey.}
    property Macros[const MKey : word] : pTMacroRec read GetKeyedMacro;

    {NamedMacros returns a pointer the individual macro pointer (see WMTypes
    for more information) with a given Name.}
    property NamedMacros[const Name : string] : pTMacroRec read GetNamedMacro;

    {IndexedMacros returns a pointer the individual macro pointer (see WMTypes
    for more information) with a given Index.}
    property IndexedMacros[const Index : Integer] : pTMacroRec read GetNumberedMacro;

    {Names returns/sets the name of a macro with a given MKey}
    property Names[const MKey : word] : string read GetMacroName write SetMacroName;

    {Keys returns/sets the MKey of a macro with a given Index}
    property Keys[const Index : Integer] : Word read GetMacroKey write SetMacroKey;

    {KeyNames returns the textual name of a given MKey}
    property KeyNames[const MKey : word] : string read GetKeyName;

    {CodeHooks returns/sets the codehook (procedure executed when the hotkey is
    pressed -- occurs in lieu of keys being played back) of a macro with a given MKey.}
    property CodeHooks[const MKey : word] : CodeHookProc read GetCodeHook write SetCodeHook;

    {Control returns/sets the status of the CtrlOn bit flag in the high order byte
    of a given MKey}
    property Control[var MKey : word] : Boolean read GetControl write SetControl;

    {Alt returns/sets the status of the AltOn bit flag in the high order byte
    of a given MKey}
    property Alt[var MKey : word] : Boolean read GetAlt write SetAlt;

    {Shift returns/sets the status of the ShiftOn bit flag in the high order byte
    of a given MKey}
    property Shift[var MKey : word] : Boolean read GetShift write SetShift;

    {ScrapMacro returns a pointer to macro engine's scrap macro}
    property ScrapMacro : pTMacroRec read GetScrapMacro;

    {MacroState returns the status of the macro engine's MacroState global variable}
    property MacroState : TMacroState read GetMacroState;

    {PossibleKeys retrieves, by index, the keys that the macro engine recognizes}
    property PossibleKeys[const Index : Word] : Word read GetPossibleKey;

    {MaxPossibleKeys returns number of possible keys that the macro engine recognizes}
    property MaxPossibleKeys : Word read GetMaxPossibleKeys;

    {StateName returns the textual name of the TUserMacroState word passed into the
    Recorded and Played event handlers.  These names are stored in a string table
    in the engine's DLL and can be changed with a resource editor}
    property StateName[const Index : TUserMacroState] : String read GetStateName;
  published

    {MaxMacros determines how many total macros may be allocated}
    property MaxMacros : word read FMaxMacros write FMaxMacros default DefaultMaxMacros;

    {MaxMacroLen determines how long a macro may be.  Keep in mind that this
    parameter reflects the number of _messages_ that can be stored, not the
    number of _keys_ or _mouse events_ that can be.  Every key pressed generates
    at least two events:  one for the key down and one for the key up.  Also,
    mouse movements tend to generate lots of messages in a hurry.  See the RecordMouse
    property for a way to limit which mouse messages are recorded.}
    property MaxMacroLen : word read FMaxMacroLen write FMaxMacroLen default DefaultMaxMacroLen;

    {RecordMouse determines which mouse messages are recorded.  Possible values are
    Ignore -- no mouse recording; Clicks and Drags -- record only mouse drags and
    clicks; and All, record all mouse messages.  Set this to Ignore or Clicks And
    Drags in order to keep mouse movement from quickly eating up all available
    macro message space.}
    property RecordMouse : TRecordMouse read FRecordMouse write FRecordMouse default DefaultRecordMouse;

    {SystemWide determines whether a macro will play back from within any application,
    or just the current one.  Set this parameter to false if you wish to use the
    macro engine merely as a command processor.}
    property SystemWide : boolean read FSystemWide write FSystemWide default DefaultSystemWide;

    {Fastplayback determines whether a macro is played back at recorded speed
    or as quickly as possible.}
    property FastPlayback : boolean read FFastPlayback write FFastPlayback default DefaultFastPlayback;

    {MouseBase determines whether mouse message pointer coordinates are based on
    the screen as a whole, the current window or the initial position of the mouse pointer.
    By setting mouse coordinates to be window relative, macros containing mouse
    messages will playback correctly to windows that have been moved.  Setting the
    MouseBase to MBPointer allows macros to be played back starting with the current
    location of the mouse pointer.  This would be handy in a paint program, for
    example, when drawing figures of different types.  You would position the mouse
    pointer where you'd like the new figure, then playback the macro you previously
    recorded responsible for "drawing" it.}
    property MouseBase : TMouseBase read FMouseBase write FMouseBase default DefaultMouseBase;

    {CancelChecking determines whether the macro engine looks for the
    cancel key (defined by MK_CANCEL) while macros are being played back or
    recorded.  Because of the unusual manner in which the cancel key must be detected,
    it cannot be a shifted key.  That is, it can't require Control, Alt or Shift in
    conjunction with some other key on the keyboard -- it must be that key alone.}
    property CancelChecking : boolean read FCancelChecking write FCancelChecking default DefaultCancelChecking;

    property ToggleRecordKey : Word read GetToggleRecordKey write SetToggleRecordKey default DefaultToggleRecordKey;
    property CancelKey : Word read GetCancelKey write SetCancelKey default DefaultCancelKey;
    property PauseKey : Word read GetPauseKey write SetPauseKey default DefaultPauseKey;

    {WINMACRO.DLL version number as a string}
    property Version : String read GetVersionString write FVersion;

    property OnRecord     : TnotifyEvent read FOnRecord write FOnRecord;
    property OnRecordStop : TnotifyEvent read FOnRecordStop write FOnRecordStop;
    property OnRecorded   : TnotifyStateEvent read FOnRecorded write FOnRecorded;
    property OnPlay       : TnotifyEvent read FOnPlay write FOnPlay;
    property OnPlayEnd    : TnotifyEvent read FOnPlayEnd write FOnPlayEnd;
    property OnPaused     : TnotifyEvent read FOnPaused write FOnPaused;
    property OnResumed    : TnotifyEvent read FOnResumed write FOnResumed;
    property OnPlayed     : TnotifyStateEvent read FOnPlayed write FOnPlayed;
    property OnAssigned   : TnotifyAssignEvent read FOnAssigned write FOnAssigned;
  end;

{TMacroKeyEdit component class

This component subclasses the TEdit component.  It traps all mouse and keyboard
messages and changes it's Text property to reflect the name of the pressed key or
button.  The left mouse button is left to its default behavior so that the user
may properly navigate with the mouse.  Were the control to also absorb the unshifted
left mouse button, selecting away from it to another control would be impossible.

If a mouse is not present, the control also enables various standard navigation
keys.  These include:  MK_TAB, MK_ENTER, MK_UP, MK_DOWN, MK_SHIFTTAB, MK_ALTUP
and MK_ALTDOWN.

When a key or button is pressed that the control translates to its corresponding
name, it also sets its MKey property to the numeric value of the pressed key.  This
property can then be used in calls to those methods of TMacroSet methods that require
an MKey.

}
TMacroKeyEdit = class(Tedit)
Private
  FMKey : Word;
  FMousePresent : Boolean;
  constructor Create(AOwner: TComponent);
  function NavKey(const MKey : Word) : Boolean;
  function GetKeyName(const InMKey : Word) : string;
  procedure SetMKey(const InMKey : Word);
  function KeyDown(var Message: TMessage) : Longint;
  function ButtonDown(var Message: TMessage) : Longint;
  function KeyUp(var Message: TMessage) : Longint;
  function ButtonUp(var Message: TMessage) : Longint;
  function CharTyped(var Message: TMessage) : Longint;
  procedure NonSysKeyDown(var Message: TMessage); message WM_KEYDOWN;
  procedure NonSysKeyUp(var Message: TMessage); message WM_KEYUP;
  procedure SysKeyDown(var Message: TMessage); message WM_SYSKEYDOWN;
  procedure SysKeyUp(var Message: TMessage); message WM_SYSKEYUP;
  procedure CharMsg(var Message: TMessage); message WM_CHAR;
  procedure SysCharMsg(var Message: TMessage); message WM_SYSCHAR;
  procedure LButtonDown(var Message: TMessage); message WM_LBUTTONDOWN;
  procedure LButtonUp(var Message: TMessage); message WM_LBUTTONUP;
  procedure LButtonDblClk(var Message: TMessage); message WM_LBUTTONDBLCLK;
  procedure RButtonDown(var Message: TMessage); message WM_RBUTTONDOWN;
  procedure RButtonUp(var Message: TMessage); message WM_RBUTTONUP;
  procedure RButtonDblClk(var Message: TMessage); message WM_RBUTTONDBLCLK;
  procedure MButtonDown(var Message: TMessage); message WM_MBUTTONDOWN;
  procedure MButtonUp(var Message: TMessage); message WM_MBUTTONUP;
  procedure MButtonDblClk(var Message: TMessage); message WM_MBUTTONDBLCLK;
public
  property MKey : Word read FMKey write SetMKey Default DefaultMKey;
end;

{TMacroKeyComboBox component class

This component subclasses the TComboBox component.  It traps all mouse and keyboard
messages and changes it's Text property to reflect the name of the pressed key or
button.  The left mouse button is left to its default behavior so that the user
may properly navigate with the mouse.  Were the control to also absorb the unshifted
left mouse button, selecting away from it to another control would be impossible.

If a mouse is not present, the control also enables various standard navigation
keys.  These include:  MK_TAB, MK_ENTER, MK_UP, MK_DOWN, MK_SHIFTTAB, MK_ALTUP
and MK_ALTDOWN.

The drop down list includes the textual name of all keys that the macro engine
recognizes.  Clicking on one of them has the same effect as pressing the key or
mouse button in the control's edit field.  Pressing a key or mouse button while
the list is dropped down will cause it to jump to the entry in the list
corresponding to the key.  In other words, it works as you would expect.

When a key or button is pressed that the control translates to its corresponding
name, it also sets its MKey property to the numeric value of the pressed key.  This
property can then be used in calls to those methods of TMacroSet methods that require
an MKey.

}
TMacroKeyComboBox = class(TComboBox)
Private
  FMaxPossibleMKeys : Word;
  FMKey : Word;
  FMousePresent : Boolean;
  function BuildPossibleMKeys : Word;
  procedure ComboWndProc(var Message: TMessage; ComboWnd: HWnd; ComboProc: Pointer); override;
  constructor Create(AOwner: TComponent); override;
  function NavKey(const InMKey : Word) : Boolean;
  function GetKeyName(const InMKey : Word) : string;
  procedure SetMKey(const InMKey : Word);
  function KeyDown(var Message: TMessage) : Longint;
  function ButtonDown(var Message: TMessage) : Longint;
  function KeyUp(var Message: TMessage) : Longint;
  function ButtonUp(var Message: TMessage) : Longint;
  function CharTyped(var Message: TMessage) : Longint;
  procedure NonSysKeyDown(var Message: TMessage); message WM_KEYDOWN;
  procedure NonSysKeyUp(var Message: TMessage); message WM_KEYUP;
  procedure SysKeyDown(var Message: TMessage); message WM_SYSKEYDOWN;
  procedure SysKeyUp(var Message: TMessage); message WM_SYSKEYUP;
  procedure CharMsg(var Message: TMessage); message WM_CHAR;
  procedure SysCharMsg(var Message: TMessage); message WM_SYSCHAR;
  procedure LButtonDown(var Message: TMessage); message WM_LBUTTONDOWN;
  procedure LButtonUp(var Message: TMessage); message WM_LBUTTONUP;
  procedure LButtonDblClk(var Message: TMessage); message WM_LBUTTONDBLCLK;
  procedure RButtonDown(var Message: TMessage); message WM_RBUTTONDOWN;
  procedure RButtonUp(var Message: TMessage); message WM_RBUTTONUP;
  procedure RButtonDblClk(var Message: TMessage); message WM_RBUTTONDBLCLK;
  procedure MButtonDown(var Message: TMessage); message WM_MBUTTONDOWN;
  procedure MButtonUp(var Message: TMessage); message WM_MBUTTONUP;
  procedure MButtonDblClk(var Message: TMessage); message WM_MBUTTONDBLCLK;
  procedure Click; override;
  procedure DropDown; override;
public
  property MKey : Word read FMKey write SetMKey Default DefaultMKey;
end;


implementation

procedure Register;
begin
  RegisterComponents('Macros',[TMacroSet, TMacroKeyEdit, TMacroKeyComboBox]);
  RegisterPropertyEditor(TypeInfo(Word), TMacroSet, 'ToggleRecordKey', TMKeyProperty);
  RegisterPropertyEditor(TypeInfo(Word), TMacroSet, 'CancelKey', TMUnshiftedKeyProperty);
  RegisterPropertyEditor(TypeInfo(Word), TMacroSet, 'PauseKey', TMUnshiftedKeyProperty);
end;

constructor TMacroSet.Create(AOwner: TComponent);
begin

  inherited Create(AOwner);
  FMacroContext:=nil;

  {Initialize fields to defaults}
  FMaxMacros:=DefaultMaxMacros;
  FMaxMacroLen:=DefaultMaxMacroLen;
  FFastPlayback:=DefaultFastPlayback;
  FSystemWide:=DefaultSystemWide;
  FRecordMouse:=DefaultRecordMouse;
  FMouseBase:=DefaultMouseBase;
  CancelKey:=DefaultCancelKey;
  PauseKey:=DefaultPauseKey;
  ToggleRecordKey:=DefaultToggleRecordKey;

  FCallingWindow := AllocateHWnd(WndProc); {Setup the message processing window}
end;

destructor TMacroSet.Destroy;
begin
  DeallocateHWnd(FCallingWindow); {Destroy the message window}
  inherited Destroy;
end;

{Translate messages received from the macro engine into TNotifyEvents}
procedure TMacroSet.WndProc(var Message: TMessage);
begin
  with TWMMacMsg(Message) do
    case Msg of
      MM_RECORDON : If Assigned(FOnRecord) then FOnRecord(Self);
      MM_RECORDOFF : If Assigned(FOnRecordStop) then FOnRecordStop(Self);
      MM_PLAYBACKON : If Assigned(FOnPlay) then FOnPlay(Self);
      MM_PLAYBACKOFF : If Assigned(FOnPlayEnd) then FOnPlayEnd(Self);
      MM_PLAYBACKPAUSED : If Assigned(FOnPaused) then FOnPaused(Self);
      MM_PLAYBACKRESUMED : If Assigned(FOnResumed) then FOnResumed(Self);
      MM_RECORDCHANGE : If Assigned(FOnRecorded) then FOnRecorded(Self,TUserMacroState(State));
      MM_PLAYBACKCHANGE : If Assigned(FOnPlayed) then FOnPlayed(Self,TUserMacroState(State));
      MM_MACROASSIGNED : If Assigned(FOnAssigned) then FOnAssigned(Self,pTMacroRec(Macro));
    else
      Result := DefWindowProc(FCallingWindow, Msg, State, Macro);
    end;
end;

{See the class method declarations above for descriptions of the following methods.}

{Method implentations for the TMacroSet component class}

function TMacroSet.GetNumberedMacro(const Index : Integer) : pTMacroRec;
begin
  Result:=GetMacroByIndex(FMacroContext,Index);
  If (Result=nil) then raise EMacroNotFound.create(LoadMsg(MGMacroNotFound));
end;

function TMacroSet.GetNamedMacro(const Name : string) : pTMacroRec;
var
  Mcnt : word;
  ScrapMacroRec : ptMacroRec;

begin
  Result:=nil;
  ScrapMacroRec:=GetScrapMacro;
  With FMacroContext^ do
    For Mcnt:=0 to pred(MaxMacros) do
      If (UpperCase(StrPas(WMImport.GetMacroName(Macros^[Mcnt])))=UpperCase(Name)) then
        Result := Macros^[Mcnt];
  If (Result=nil) and (ScrapMacroRec^.MHDR.Name<>nil)
  and (UpperCase(StrPas(ScrapMacroRec^.MHDR.Name))=UpperCase(Name)) then Result:=ScrapMacroRec;
  If (Result=nil) then raise EMacroNotFound.create(LoadMsg(MGMacroNotFound));
end;

function TMacroSet.GetKeyedMacro(const MKey : Word) : pTMacroRec;
begin
  Result:=GetMacroByKey(FMacroContext,MKey);
  If (Result=nil) then raise EMacroNotFound.create(LoadMsg(MGMacroNotFound));
end;

function TMacroSet.GetMacroName(const MKey : Word) : string;
var
  MRec : pTMacroRec;
begin
  Result:='';
  MRec := GetMacroByKey(FMacroContext,MKey);
  If (MRec<>nil) then
    Result:=StrPas(WMImport.GetMacroName(MRec))
  else raise EMacroNotFound.create(LoadMsg(MGMacroNotFound));
end;

procedure TMacroSet.SetMacroName(const MKey : Word; const Name : String);
var
  MRec : pTMacroRec;
  Buffer : array[0..MacroNameStringLen] of char;
begin
  MRec := GetMacroByKey(FMacroContext,MKey);
  If (MRec<>nil) then
    WMImport.SetMacroName(MRec, StrPCopy(Buffer,Name))
  else raise EMacroNotFound.create(LoadMsg(MGMacroNotFound));
end;

function TMacroSet.GetMacroKey(const Index : Integer) : Word;
Var
  MRec : pTMacroRec;
begin
  Result:=0;
  MRec := GetMacroByIndex(FMacroContext,Index);
  If (MRec<>nil) then
    Result:=WMImport.GetMacroKey(MRec)
  else raise EMacroNotFound.create(LoadMsg(MGMacroNotFound));
end;

procedure TMacroSet.SetMacroKey(const Index : Integer; const MKey : Word);
Var
  MRec : pTMacroRec;
begin
  MRec := GetMacroByIndex(FMacroContext,Index);
  If (MRec<>nil) then
    WMImport.SetMacroKey(MRec, MKey)
  else raise EMacroNotFound.create(LoadMsg(MGMacroNotFound));
end;

function TMacroSet.GetKeyName(const MKey : Word) : string;
const
  MaxTextLen =  30;
var
  Buffer : array[0..MaxTextLen] of char;
begin
  Result:=StrPas(WMImport.GetKeyName(MKey, Buffer, MaxTextLen));
end;

function TMacroSet.GetCodeHook(const MKey : Word) : CodeHookProc;
var
  MRec : pTMacroRec;
begin
  Result:=nil;
  MRec:=GetMacroByKey(FMacroContext,MKey);
  If (MRec<>nil) then
    Result:=GetMacroCodeHook(Mrec)
  else raise EMacroNotFound.create(LoadMsg(MGMacroNotFound));
end;

procedure TMacroSet.SetCodeHook(const MKey : Word; const CodeHook : CodeHookProc);
var
  MRec : pTMacroRec;
begin
  MRec:=GetMacroByKey(FMacroContext,MKey);
  If (MRec<>nil) then
    SetMacroCodeHook(MRec, CodeHook)
  else raise EMacroNotFound.create(LoadMsg(MGMacroNotFound));
end;

function TMacroSet.GetPossibleKey(const Index : Word) : word;
begin
  Result:=WMImport.GetPossibleMKey(Index);
end;

function TMacroSet.GetMaxPossibleKeys : Word;
begin
  Result:=WMImport.GetMaxPossibleMKeys;
end;

function TMacroSet.GetControl(var MKey : word) : boolean;
begin
  Result:=WMImport.GetKeyCtrl(MKey);
end;

procedure TMacroSet.SetControl(var MKey : Word; const Ctrl : boolean);
begin
  WMImport.SetKeyCtrl(MKey, Ctrl);
end;

function TMacroSet.GetAlt(var MKey : word) : boolean;
begin
  Result:=WMImport.GetKeyAlt(MKey);
end;

procedure TMacroSet.SetAlt(var MKey : Word; const Alt : boolean);
begin
  WMImport.SetKeyAlt(MKey,Alt);
end;

function TMacroSet.GetShift(var MKey : word) : boolean;
begin
  Result:=WMImport.GetKeyShift(MKey);
end;

procedure TMacroSet.SetShift(var MKey : Word; const Shift : boolean);
begin
  WMImport.SetKeyShift(MKey,Shift);
end;

function TMacroSet.GetScrapMacro : pTMacroRec;
begin
  Result:=WMImport.GetScrapMacro;
end;

procedure TMacroSet.Add(const MKey : word;
                              const Macro : pTMacroRec;
                              const Name : String;
                              const CodeHook : CodeHookProc);
var
  Buffer : array[0..MacroNameStringLen] of char;
begin
  If (WMImport.InstallMacro(FMacroContext,MKey,Macro,StrPCopy(Buffer,Name),CodeHook)>MRSevereError) then
    raise EInstallMacro.create(LoadMsg(MGAllocateMacro));
end;

procedure TMacroSet.Delete(const MKey : word);
begin
  If (WMImport.GetMacroByKey(FMacroContext,MKey)<>nil) then begin
    If (WMImport.InstallMacro(FMacroContext,MKey,nil,'',nil)>MRSevereError) then
      raise EInstallMacro.create(LoadMsg(MGDeleteMacro));
  end else raise EMacroNotFound.create(LoadMsg(MGMacroNotFound));
end;

procedure TMacroSet.ToggleRecording;
begin
  WMImport.ToggleRecording;
end;

procedure TMacroSet.SetRecordingOn;
begin
  WMImport.RecordingOn;
end;

procedure TMacroSet.SetRecordingOff;
begin
  WMImport.RecordingOff;
end;

function TMacroSet.GetMacroState : TMacroState;
begin
  Result:=WMImport.GetMacroState;
end;

function TMacroSet.GetToggleRecordKey : Word;
begin
  Result:=WMImport.GetToggleRecordKey;
end;

procedure TMacroSet.SetToggleRecordKey(const MKey : Word);
begin
  WMImport.SetToggleRecordKey(MKey);
end;

function TMacroSet.GetCancelKey : Word;
begin
  Result:=WMImport.GetCancelKey;
end;

procedure TMacroSet.SetCancelKey(const MKey : Word);
begin
  If (not(WMImport.SetCancelKey(MKey))) then
    raise EShiftKeysNotAllowed.Create(LoadMsg(MGShiftKeysNotAllowed));
end;

function TMacroSet.GetPauseKey : Word;
begin
  Result:=WMImport.GetPauseKey;
end;

procedure TMacroSet.SetPauseKey(const MKey : Word);
begin
  If (not(WMImport.SetPauseKey(MKey))) then
    raise EShiftKeysNotAllowed.Create(LoadMsg(MGShiftKeysNotAllowed));
end;

procedure TMacroSet.PlaybackMacro(const MKey : Word);
var
  MRec : pTMacroRec;
begin
  MRec := GetMacroByKey(FMacroContext,MKey);
  If (MRec<>nil) then WMImport.PlaybackMacro(Mrec)
  else raise EMacroNotFound.create(LoadMsg(MGMacroNotFound));
end;

procedure TMacroSet.TypeKeys(const KeyChars : PChar);
begin
  WMImport.TypeKeys(KeyChars);
end;

procedure TMacroSet.PlaybackScrap;
begin
  WMImport.PlaybackScrap;
end;

procedure TMacroSet.SaveToFile(const FileName : string);
var
  Buffer : array[0..Pred(SizeOf(String))] of Char;
begin
  If (WMImport.WriteMacros(FMacroContext,StrPCopy(Buffer,FileName))<>0) then
    raise EWriteMacros.create(LoadMsg(MGWriteMacroFile));
end;

procedure TMacroSet.ReadFromFile(const FileName : string; const Merge : boolean);
var
  Buffer : array[0..Pred(SizeOf(String))] of Char;
begin
  If (WMImport.ReadMacros(FMacroContext,StrPCopy(Buffer,FileName), Merge)<>0) then
    raise EReadMacros.create(LoadMsg(MGReadMacroFile));
end;

function TMacroSet.MacrosOn: boolean;
{MacrosOn

This function has a dual purpose:  1) Install the macro handler,
2) Act as a function that tells whether or not the handler has been
installed.  This "multiple-personality" approach allows the user to use just
one function to turn macro handling on and to later determine whether it
has been turned on (prior to turning it off, for example).  Since the
function result can be discarded, and the function called as though it
were a procedure, the user uses

MacrosOn;

to turn handling on.  To determine whether handling is installed (e.g., to
turn it off)

If MacrosOn then MacrosOff;

will do just fine.
}
begin
  If (FMacroContext=nil) then begin {We're not already On}
     FMacroContext:=InstallHook(FCallingWindow,FMaxMacros,FMaxMacroLen,FFastPlayback,
                                FSystemWide,FRecordMouse,FMouseBase,FCancelChecking);
     If (FMacroContext=nil) then raise EHookError.create(LoadMsg(MGAllocateContext));
  end;
  Result:=(FMacroContext<>nil);
end;

procedure TMacroSet.MacrosOff;
begin
  UninstallHook(FMacroContext);
end;

function TMacroSet.GetStateName(const Index : TUserMacroState) : string;
begin
  Result:=LoadMsg(SNNone+Byte(Index));
end;

function TMacroSet.GetVersionString : String;
var
  Temp : Word;
begin
  Temp:=WMImport.GetWinmacroVersion;
  Result:=IntToStr(Hi(Temp))+'.'+IntToStr(Lo(Temp));
end;

{Method implemenations for the TMKeyProperty property editor class}

function TMKeyProperty.GetKeyName(const MKey : Word) : string;
const
  MaxTextLen =  30;
var
  Buffer : array[0..MaxTextLen] of char;
begin
  Result:=StrPas(WMImport.GetKeyName(MKey, Buffer, MaxTextLen));
end;

function TMKeyProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paValueList];
end;

function TMKeyProperty.GetValue: string;
begin
    Result:=GetKeyName(GetOrdValue);
end;

procedure TMKeyProperty.GetValues(Proc: TGetStrProc);
var
  MCnt, MaxKeys : Word;
begin
  MaxKeys:=pred(WMImport.GetMaxPossibleMKeys);
  For MCnt:=0 to MaxKeys do
    Proc(GetKeyName(WMImport.GetPossibleMKey(Mcnt)));
end;

procedure TMKeyProperty.SetValue(const Value: string);
var
  MCnt, MaxKeys : Word;
begin
    MaxKeys:=pred(WMImport.GetMaxPossibleMKeys);
    For MCnt:=0 to MaxKeys do
      if (CompareText(GetKeyName(WMImport.GetPossibleMKey(Mcnt)), Value) = 0) then begin
        SetOrdValue(WMImport.GetPossibleMKey(Mcnt));
        Exit;
      end;
    raise EPropertyError.Create(LoadMsg(MGInvalidPropertyValue));
end;

{Method implemenations for the TMUnshiftedKeyProperty property editor class}

function TMUnshiftedKeyProperty.GetKeyName(const MKey : Word) : string;
const
  MaxTextLen =  30;
var
  Buffer : array[0..MaxTextLen] of char;
begin
  Result:=StrPas(WMImport.GetKeyName(MKey, Buffer, MaxTextLen));
end;

function TMUnshiftedKeyProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paValueList];
end;

function TMUnshiftedKeyProperty.GetValue: string;
begin
    Result:=GetKeyName(GetOrdValue);
end;

procedure TMUnshiftedKeyProperty.GetValues(Proc: TGetStrProc);
var
  MCnt, MaxKeys, MKey : Word;
begin
  MaxKeys:=pred(WMImport.GetMaxPossibleMKeys);
  For MCnt:=0 to MaxKeys do begin
    MKey:=WMImport.GetPossibleMKey(Mcnt);
    If (Hi(MKey)=0) then Proc(GetKeyName(MKey));
  end;
end;

procedure TMUnshiftedKeyProperty.SetValue(const Value: string);
var
  MCnt, MaxKeys, MKey : Word;
begin
    MaxKeys:=pred(WMImport.GetMaxPossibleMKeys);
    For MCnt:=0 to MaxKeys do begin
      MKey:=WMImport.GetPossibleMKey(Mcnt);
      if (Hi(MKey)=0) and (CompareText(GetKeyName(MKey), Value) = 0) then begin
        SetOrdValue(WMImport.GetPossibleMKey(Mcnt));
        Exit;
      end;
    end;
    raise EPropertyError.Create(LoadMsg(MGInvalidPropertyValue));
end;

constructor TMacroKeyEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FMKey:=0;
  FMousePresent:=ByteBool(GetSystemMetrics(SM_MOUSEPRESENT));
end;

function TMacroKeyEdit.NavKey(const MKey : Word) : Boolean;
begin
  Case MKey of
    MK_TAB, MK_ENTER, MK_UP, MK_DOWN, MK_SHIFTTAB, MK_ALTUP, MK_ALTDOWN : Result:=true;
  else Result:=false;
  end;
end;

function TMacroKeyEdit.KeyDown(var Message: TMessage) : Longint;
begin
  FMKey:=Message.wparam;
  Text:=GetKeyName(FMKey);
  Result:=1;
end;

function TMacroKeyEdit.KeyUp(var Message: TMessage) : Longint;
begin
  Result:=1;
end;

function TMacroKeyEdit.ButtonDown(var Message: TMessage) : Longint;
begin
  With Message do begin
    Wparam:=MessageToKey[Msg].VKey;
    WMImport.GetShiftStates(WParam);
  end;
  KeyDown(Message);
  Result:=1;
end;

function TMacroKeyEdit.ButtonUp(var Message: TMessage) : Longint;
begin
  Result:=1;
end;

function TMacroKeyEdit.CharTyped(var Message: TMessage) : Longint;
begin
  Result:=1;
end;

procedure TMacroKeyEdit.NonSysKeyDown(var Message: TMessage);
Var
  MKey : Word;
begin
  With Message do begin
    MKey:=wparam;
    WMImport.GetShiftStates(MKey);
    If ((not(FMousePresent)) and (NavKey(MKey))) or (MKey in ShiftVKeys) then exit
    else begin
      wparam:=MKey;
      Result:=KeyDown(Message);
    end;
  end;
end;

procedure TMacroKeyEdit.NonSysKeyUp(var Message: TMessage);
Var
  MKey : Word;
begin
  With Message do begin
    MKey:=wparam;
    WMImport.GetShiftStates(MKey);
    If ((not(FMousePresent)) and (NavKey(MKey))) or (MKey in ShiftVKeys) then exit
    else begin
      wparam:=MKey;
      Result:=KeyUp(Message);
    end;
  end;
end;

procedure TMacroKeyEdit.SysKeyDown(var Message: TMessage);
begin
  NonSysKeyDown(Message);
end;

procedure TMacroKeyEdit.SysKeyUp(var Message: TMessage);
begin
  NonSysKeyUp(Message);
end;

procedure TMacroKeyEdit.CharMsg(var Message: TMessage);
begin
  Message.Result:=CharTyped(Message);
end;

procedure TMacroKeyEdit.SysCharMsg(var Message: TMessage);
begin
  CharMsg(Message);
end;

procedure TMacroKeyEdit.LButtonDown(var Message: TMessage);
var
  MKey : Word;
begin
  MKey:=0;
  WMImport.GetShiftStates(MKey);
  If (MKey<>0) then Message.Result:=ButtonDown(Message)
  else Inherited;
end;

procedure TMacroKeyEdit.LButtonUp(var Message: TMessage);
var
  MKey : Word;
begin
  MKey:=0;
  WMImport.GetShiftStates(MKey);
  If (MKey<>0) then Message.Result:=ButtonUp(Message)
  else Inherited;
end;

procedure TMacroKeyEdit.LButtonDblClk(var Message: TMessage);
begin
  LButtonDown(Message);
end;

procedure TMacroKeyEdit.RButtonDown(var Message: TMessage);
begin
  Message.Result:=ButtonDown(Message);
end;

procedure TMacroKeyEdit.RButtonUp(var Message: TMessage);
begin
  Message.Result:=ButtonUp(Message);
end;

procedure TMacroKeyEdit.RButtonDblClk(var Message: TMessage);
begin
  Message.Result:=ButtonDown(Message);
end;

procedure TMacroKeyEdit.MButtonDown(var Message: TMessage);
begin
  Message.Result:=ButtonDown(Message);
end;

procedure TMacroKeyEdit.MButtonUp(var Message: TMessage);
begin
  Message.Result:=ButtonUp(Message);
end;

procedure TMacroKeyEdit.MButtonDblClk(var Message: TMessage);
begin
  Message.Result:=ButtonDown(Message);
end;

function TMacroKeyEdit.GetKeyName(const InMKey : Word) : string;
const
  MaxTextLen =  30;
var
  Buffer : array[0..MaxTextLen] of char;
begin
  Result:=StrPas(WMImport.GetKeyName(InMKey, Buffer, MaxTextLen));
end;

procedure TMacroKeyEdit.SetMKey(const InMKey : Word);
Var
  Name : String;
begin
  Name := GetKeyName(InMKey);
  If (Name<>'') then begin
    FMKey:=InMKey;
    Text:=Name;
  end;
end;

function TMacroKeyComboBox.BuildPossibleMKeys : Word;
var
  MCnt, MaxKeys : Word;
begin
    Cursor:=crHourGlass;
    Items.Clear;
    MaxKeys:=pred(WMImport.GetMaxPossibleMKeys);
    For MCnt:=0 to MaxKeys do
      Items.Add(GetKeyName(WMImport.GetPossibleMKey(Mcnt)));
    Cursor:=crDefault;
    Result:=MCnt;
end;

constructor TMacroKeyComboBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FMaxPossibleMKeys:=0;
  FMKey:=0;
  FMousePresent:=ByteBool(GetSystemMetrics(SM_MOUSEPRESENT));
end;

function TMacroKeyComboBox.GetKeyName(const InMKey : Word) : string;
const
  MaxTextLen =  30;
var
  Buffer : array[0..MaxTextLen] of char;
begin
  Result:=StrPas(WMImport.GetKeyName(InMKey, Buffer, MaxTextLen));
end;

procedure TMacroKeyComboBox.SetMKey(const InMKey : Word);
var
  Ind : Integer;
begin
   If (FMaxPossibleMKeys=0) then FMaxPossibleMKeys:=BuildPossibleMKeys;
  Ind := WMImport.GetPossibleMKeyIndex(InMKey);
  If (Ind<>-1) then begin
    FMKey:=InMKey;
    Text:=GetKeyName(FMKey);
    If (Items.Count<>0) then ItemIndex:=Ind;
    inherited Change;
  end;
end;

function TMacroKeyComboBox.KeyDown(var Message: TMessage) : Longint;
begin
    If (FMaxPossibleMKeys=0) then FMaxPossibleMKeys:=BuildPossibleMKeys;
    With Message do begin
        FMKey:=wparam;
        Text:= GetKeyName(FMKey);
	 ItemIndex := WMImport.GetPossibleMKeyIndex(FMKey);
        inherited Change;
    end;
    Result:=1;
end;

function TMacroKeyComboBox.KeyUp(var Message: TMessage) : Longint;
begin
  Result:= 1;
end;

function TMacroKeyComboBox.ButtonDown(var Message: TMessage) : Longint;
begin
    With Message do Wparam:=MessageToKey[Msg].VKey;
    NonSysKeyDown(Message);
    Result:=1;
end;

function TMacroKeyComboBox.ButtonUp(var Message: TMessage) : Longint;
begin
  Result:= 1;
end;

function TMacroKeyComboBox.CharTyped(var Message: TMessage) : Longint;
begin
  Result:= 1;
end;

procedure TMacroKeyComboBox.NonSysKeyDown(var Message: TMessage);
Var
  InMKey : Word;
begin
  With Message do begin
    InMKey:=wparam;
    WMImport.GetShiftStates(InMKey);
    If ((not(FMousePresent)) and (NavKey(InMKey))) or (InMKey in ShiftVKeys) then exit
    else begin
      wparam:=InMKey;
      Result:=KeyDown(Message);
    end;
  end;
end;

procedure TMacroKeyComboBox.NonSysKeyUp(var Message: TMessage);
Var
  InMKey : Word;
begin
  With Message do begin
    InMKey:=wparam;
    WMImport.GetShiftStates(InMKey);
    If ((not(FMousePresent)) and (NavKey(InMKey))) or (InMKey in ShiftVKeys) then exit
    else begin
      wparam:=InMKey;
      Result:=KeyUp(Message);
    end;
  end;
end;

procedure TMacroKeyComboBox.SysKeyDown(var Message: TMessage);
begin
  NonSysKeyDown(Message);
end;

procedure TMacroKeyComboBox.SysKeyUp(var Message: TMessage);
begin
  NonSysKeyUp(Message);
end;

procedure TMacroKeyComboBox.CharMsg(var Message: TMessage);
begin
  Message.Result:=CharTyped(Message);
end;

procedure TMacroKeyComboBox.SysCharMsg(var Message: TMessage);
begin
  CharMsg(Message);
end;

procedure TMacroKeyComboBox.LButtonDown(var Message: TMessage);
var
  InMKey : Word;
begin
  InMKey:=0;
  WMImport.GetShiftStates(InMKey);
  If (InMKey<>0) then Message.Result:=ButtonDown(Message)
  else Inherited;
end;

procedure TMacroKeyComboBox.LButtonUp(var Message: TMessage);
var
  InMKey : Word;
begin
  InMKey:=0;
  WMImport.GetShiftStates(InMKey);
  If (InMKey<>0) then Message.Result:=ButtonUp(Message)
  else Inherited;
end;

procedure TMacroKeyComboBox.LButtonDblClk(var Message: TMessage);
begin
  LButtonDown(Message);
end;

procedure TMacroKeyComboBox.RButtonDown(var Message: TMessage);
begin
  Message.Result:=ButtonDown(Message);
end;

procedure TMacroKeyComboBox.RButtonUp(var Message: TMessage);
begin
  Message.Result:=ButtonUp(Message);
end;

procedure TMacroKeyComboBox.RButtonDblClk(var Message: TMessage);
begin
  Message.Result:=ButtonDown(Message);
end;

procedure TMacroKeyComboBox.MButtonDown(var Message: TMessage);
begin
  Message.Result:=ButtonDown(Message);
end;

procedure TMacroKeyComboBox.MButtonUp(var Message: TMessage);
begin
  Message.Result:=ButtonUp(Message);
end;

procedure TMacroKeyComboBox.MButtonDblClk(var Message: TMessage);
begin
  Message.Result:=ButtonDown(Message);
end;

procedure TMacroKeyComboBox.ComboWndProc(var Message: TMessage; ComboWnd: HWnd;
  ComboProc: Pointer);
Var
  InMKey : Word;
begin
  with Message do
  begin
    case Msg of
      WM_KEYDOWN, WM_SYSKEYDOWN: begin
          if DoKeyDown(TWMKey(Message)) then Exit;
          NonSysKeyDown(Message);
          exit;
        end;
      WM_CHAR, WM_SYSCHAR: begin
          if DoKeyPress(TWMKey(Message)) then Exit;
          Message.Result:=CharTyped(Message);
          exit;
        end;
      WM_KEYUP, WM_SYSKEYUP: begin
          if DoKeyUp(TWMKey(Message)) then Exit;
          NonSysKeyUp(Message);
          exit;
        end;
      WM_LBUTTONDOWN : begin
          InMKey:=0;
          WMImport.GetShiftStates(InMKey);
          If (InMKey<>0) then begin
            Message.Result:=ButtonDown(Message);
            exit;
          end;
        end;
      WM_RBUTTONDOWN, WM_MBUTTONDOWN : begin
            Message.Result:=ButtonDown(Message);
            exit;
        end;
      WM_LBUTTONUP : begin
          InMKey:=0;
          WMImport.GetShiftStates(InMKey);
          If (InMKey<>0) then begin
            Message.Result:=ButtonUp(Message);
            exit;
          end;
        end;
      WM_RBUTTONUP, WM_MBUTTONUP : begin
            Message.Result:=ButtonUp(Message);
            exit;
        end;
      WM_LBUTTONDBLCLK, WM_RBUTTONDBLCLK, WM_MBUTTONDBLCLK : begin
          Message.Result:=ButtonDown(Message);
          exit;
        end;
      WM_SETFOCUS:
        GetParentForm(Self).Perform(CM_FOCUSCHANGED, 0, Longint(Self));
      WM_NCHITTEST:
        if csDesigning in ComponentState then
        begin
          Result := HTTRANSPARENT;
          Exit;
        end;
      CN_KEYDOWN, CN_CHAR, CN_SYSKEYDOWN, CN_SYSCHAR:
        begin
          WndProc(Message);
          Exit;
        end;
    end;
    Result := CallWindowProc(ComboProc, ComboWnd, Msg, WParam, LParam);
  end;
end;

procedure TMacroKeyComboBox.Click;
var
  Index : Word;
begin
  If (FMaxPossibleMKeys=0) then FMaxPossibleMKeys:=BuildPossibleMKeys;
  Index:=SendMessage(Handle,CB_GETCURSEL,0,0);
  If (Index<>CB_ERR) then begin
    FMKey:=GetPossibleMKey(Index);
  end;
  inherited Click;
end;

function TMacroKeyComboBox.NavKey(const InMKey : Word) : Boolean;
begin
  Case InMKey of
    MK_TAB, MK_ENTER, MK_UP, MK_DOWN, MK_SHIFTTAB, MK_ALTUP, MK_ALTDOWN : Result:=true;
  else Result:=false;
  end;
end;

procedure TMacroKeyComboBox.DropDown;
begin
  If (FMaxPossibleMKeys=0) then FMaxPossibleMKeys:=BuildPossibleMKeys;
  inherited DropDown;
end;

end.

