{$I WINMACRO.INC}
{
WMTypes.pas
Interface and import unit for WINMACRO.DLL, a keyboard/mouse/code macro engine
for Windows

Written by Ken Henderson.

Copyright (c) 1994 Ken Henderson.  All Rights Reserved.

Last modification: 11-22-94

WMTypes acts as a front end to the facilities in Winmacro.dll.  It establishes
all global types and constants.  WMImport functions as the import unit for
WINMACRO.

The Internals

The macro engine stores information pertinent to a given macro "session" in what
is know as a macro "context".  Things like the maximum number of macros allowed
and what mouse events the macro engine is to record are stored in the macro context.

Each macro context contains a pointer to an array of macro handles.  These macro
handles point to the macro records that contain the recorded messages.  The
actual memory allocated to this array is sized according to the MaxMacNum parameter
passed into the InstallHook routine.

Each macro record is composed of two parts:  the header and the message sections.  The
header contains all information that is pertinent to an individual macro with the
exception of the actual recorded messages.  Things like the macro's length, its name,
the status of the toggle keys at the time the macro was recorded are all stored
in the header portion of the macro.  The message portion consists of an array of
TEventMsg's, the message format used by Windows' journaling system.  Winmacro
performs its macro magic using this system.  The message array is sized according
to the MaxMacLen parameter passed into the InstallHook routine.

A question could be asked as to why I used the journaling facility in Windows, as
opposed to recording and playing back actual keyboard and mouse messages.  The
reason for this is simple:  Windows refuses to acknowledge shift key states properly
when using that technique.  That is, posting a WM_KEYDOWN for the VK_SHIFT key,
followed by a WM_KEYDOWN for a VK_X, then following both messages with their
respective WM_KEYUP messages, does not cause a WM_CHAR message to be generated
that cooresponds to the shifted version of VK_X.  The WM_CHAR message that is
generated depends entirely upong the _physical_ keyboard at the time the VK_X key
down message is seen.  Hold down shift while the message is posted and you get a
shifted 'x'.  Short of that, I could not get Windows to properly recognize the
desired messages.  Windows apparently looks directly at the keyboard for such
information.  Hence, my use of the journaling system.

The MKey portion of the macro header contains the actual hotkey that is used to
invoke the macro.  It consists of a word devided into two parts: the primary key
and the shift key sections, respectively.  The primary key portion occupies the low
order byte of the word and is just that: the Windows virtual key code of the
key on the keyboard primarily responsible for invoking the macro.  By
"primarily responsible" I mean that this is the key that is pressed,
possibly in conjunction with shift keys, to execute the macro.  Since there are a
maximum of 255 virtual keys, a byte is sufficient.  The high order byte is
occupied by the shift states portion of the MKey.  It stores the status of the shift
keys (the Control, Alt and Shift keys) that are to be pressed in conjunction with
the primary key.  Windows' VK_... constants can be used to reference specific
keys on the keyboard.  Note that mouse buttons are also usable within MKeys,
you need only specify the appropriate virtual key code.  This includes the
ability to combine those buttons with shift keys (e.g. Control+LeftClick) to
derive hotkeys.  Note also that "normal" keys -- keys other than function or
shifted keys can be used as MKeys, as well.  You could, for example, redefine
the tilde key by setting it up as an MKey and playing back a recorded sequence
that typed some other character.

Winmacro installs a total of four Windows "hooks" at various times:  a keyboard
hook and a mouse hook when InstallHook is called, a journal record hook and a
journal playback hook when a macro is being recorded or played back.  Only
one macro may be recorded or played back at any given time.

All recording goes by default to the scrap macro.  The scrap macro's purpose is
to act as a scratch area for the engine, but it may be used by the user as well.
Only one scrap macro is defined for all macro contexts to use, hence the limitation
that only one macro may be recorded at any time.  This design is due to Windows
own limitation that journal record and playback hooks must be system wide.  Since
the InstallHook routine actually sets the maximum size a macro may be, the question
could be raised: How does the scrap macro handle recording a macro of any size?
How can it know about future calls to InstallHook that may specify macro length
limits that exceed its allocated size?  The answer is that whenever InstallHook
is called, the MaxMacLen parameter is checked against the one used to allocate
the scrap macro.  If it is larger, the scrap macro is destroyed and reallocated
using the new size.  The scrap macro never shrinks.  It only grows to accomodate
longer macro lengths as necessary.

Only one macro context can be allocated per task.  When a key or mouse event occurs,
macro contexts are searched using the current task handle as a key.  The first
match is returned.  Allocating two macros that were assigned to the same key
within two separate contexts in the same task would cause the second of the
two to never be located for playback.  It would be useless.  There should never
be a need to allocate two macro contexts within the same task.

Though only one context is allowed per task, up to ten different tasks my allocate
macro contexts and make use of the macro engine simultaneously.  If two tasks
define the same hotkey, but have both set the SystemWide parameter to false on
their respective calls to InstallHook, their hotkeys will work as expected within
each of their tasks.  If both have set SystemWide to true, the one first installed
will receive preference.  If one is SystemWide and the other is not, the SystemWide
context will receive preference across the system except within the task that is
host to the second context.  Within that task, it's own allocated context receives
preference.  In other words, it works as you would expect.

Because of the ability to limit macro playback scope to the current task, and
because of the engine's ability to execute code macros, developers can use
the macro engine to establish accelerator keys at runtime rather than just
at design time.  The keyboard can be configured while the application is
running, using the macro engine as a glorified command processor.

Three default keys are defined by Winmacro, MK_TOGGLERECORD, MK_PLAYBACKSCRAP and
MK_PAUSE.  These default to the F11, F12 and Pause keys, respectively.

Messages

When a status change occurs in the macro engine, it sends messages to the window
identified by the CallingWindow property of each macro context.  Two messages
are sent for each event that occurs:  MM_RECORDON is sent when recording is
turned on, MM_RECORDOFF is sent when it is turned off.  Neither of the messages
have any parameters.  MM_RECORDCHANGED accompanies both of them, and carries
with it the new status of the record flag.  The developer can choose to use
the message that best meets his needs.  Likewise, MM_PLAYBACKON and MM_PLAYBACKOFF
are sent when macro playback is respectively started and stopped.  MM_PLAYBACKCHANGED
accompanies both of them.

When a macro assignment occurs (from within InstallMacro), the engine sends a
MM_MACROASSIGNED message to the CallingWindow to notify it that the assignment
has happened.  This could be used to repopulate a list box listing the current
macros, for example.  The lparam field of the message is a pTMacroRec pointing
to the newly allocated macro.

}
unit WMTypes;

interface

Uses WinTypes, SysUtils, Messages;
const
  MaxMacroContext = 9;

  {Messages used by the system}

  MM_MACROFIRST = WM_USER+500;
  MM_STARTRECORD = MM_MACROFIRST;

  {I consider this method -- that of adding just 1 to each successive
   message, with each message dependent on the previous one for its
   index -- preferable because adding or deleting a message only requires
   the changing of the one immediately following the message, not all
   of the following messages.}

  MM_STOPRECORD = MM_STARTRECORD + 1;
  MM_STARTPAUSE = MM_STOPRECORD + 1;
  MM_STOPPAUSE = MM_STARTPAUSE + 1;
  MM_STARTPLAYBACK = MM_STOPPAUSE + 1;
  MM_STOPPLAYBACK = MM_STARTPLAYBACK + 1;
  MM_PAUSEPLAYBACK = MM_STOPPLAYBACK + 1;  {Necessitated by SysModal dialogs}
  MM_RESUMEPLAYBACK = MM_PAUSEPLAYBACK + 1;{Companion to MM_PAUSEPLAYBACK}

  {Messages sent to the calling task}

  MM_RECORDON = MM_RESUMEPLAYBACK + 1;
  MM_RECORDOFF = MM_RECORDON + 1;
  MM_PLAYBACKON = MM_RECORDOFF + 1;
  MM_PLAYBACKOFF = MM_PLAYBACKON + 1;
  MM_PLAYBACKPAUSED = MM_PLAYBACKOFF + 1;
  MM_PLAYBACKRESUMED = MM_PLAYBACKPAUSED + 1;
  MM_RECORDCHANGE = MM_PLAYBACKRESUMED + 1;
  MM_PLAYBACKCHANGE = MM_RECORDCHANGE + 1;
  MM_MACROASSIGNED = MM_PLAYBACKCHANGE + 1;
  MM_MACROLAST = MM_MACROASSIGNED + 1;

const

  MacroNameStringLen = 60;
  PathStrLen = 128;

  {The following is used by the macro file read/write routines}

  UnKnownIOError = $FF;

  {Messages -- implemented as a Windows string table}

   {$I WMMSGS.INC}

  {Return codes}

  MRSuccess     : byte = 0;
  MRNoMacroRoom : byte = 1;

  {MRSevereError allows us to easily distinguish between severe and non-
  severe errors.  It serves as a marker dividing non-severe errors from
  severe ones.  Errors above MRSevereError are severe errors, those under
  it aren't.  The actual code returned may or may not be of interest
  to the programmer and he can decide whether to do something specific to
  a given return code or based solely on whether a severe error occurred. }

  MRSevereError : byte = $80;
  MROutOfMemory : byte = $81;


  {MKey constant definitions for all possible keys that the macro engine
  recognizes.  Constants follow the form of MK_XXXX where XXXXX is some
  combination of 'Ctrl', 'Alt' and 'Shift' and an unshifted key name, depending on
  the key.  For example, Control+Alt+Shift+F10 is setup as MK_CtrlAltShiftF10.
  These constants can be used for the MKey parameter when application installs
  new macros.}

  {$I WMKEYS.INC}

type

  {Used to typecast word and longint variables as arrays so that their
  component bytes may be manipulated.  I realize that absoluting such variables
  requires less program code, but I consider that a non-portable technique since
  the size of word or longint variables could change in the future}

  WBytes = array[0..pred(SizeOf(Word))] of Byte;
  LBytes = array[0..Pred(SizeOf(Longint))] of Byte;

  {Current state of the macro engine}

  TMacroState = (MSNone, MSRecord, MSPlayback, MSPauseRecord);

  {Used to tell the user the current recording/playback status}

  TUserMacroState = (UMSNone, RSOff, RSOn, RSPaused, RSResumed, PSOff, PSOn, PSPaused, PSResumed);

  {Used by the engine to determine what types of mouse events to record}

  TRecordMouse = (RMIgnore, RMClicksAndDrags, RMAll);

  {Used to set the basis for mouse actions.  They can correspond to the screen
  as a whole, to the current window or to the initial position of the mouse pointer}

  TMouseBase = (MBScreen, MBWindow, MBPointer);

  {Windows' journaling facility uses TEventMsg's (despite SDK documentation
   for the journal record callback function to the contrary)}

  pTEventMsg = ^TEventMsg;
  MacroNameString = string[MacroNameStringLen];
  PathStr = string[PathStrLen];

  {The following is a necessitated by the compiler's confusion over using
   the procedure keyword in certain instances where a type derived from it
   is allowed}

  CodeHookProc = procedure;

  {TMHDR's form the first of two fields in TMacroRec's.  It might seem
  pointless to group these fields into their own record, then use the
  record in yet another record, but consider:  Memory is allocated
  to TMacroRec's based on the actual length of the macro (usually the scrap
  macro) being copied to the new macro.  When this happens, having the fields
  that never vary in size in a separate record struct allows the memory needed
  to be easily calculated using the following formula:

  MemNeeded:=sizeof(TMHDR)+(Length*sizeof(TEventMsg));

  where Length is the Length field from the source macro.  Sizeof(TEventMsg)
  is multiplied by Length because the Msgs field of a TMacroRec is an array
  of TEventMsg's.  It's this convention that allows the macros to vary in
  size without resorting to a linked list.  By grouping all but the variant
  field into a separate record structure, and then using that sub-record
  in all calculations for the "parent" record, we leave open the possibility
  that fields could be added or deleted from the TMacroRec (via the TMHDR,
  of course) without changing the calculation forumla.  The only other way
  to accomplish this flexibility would be to place "tag" variables in the
  record structure, between which all fields insertions or deletions would
  have to be made.  This solution is much more elegant.  The only negative
  to this solution is that the subrecord must proceed any reference to the fields
  that have been placed within it (of course), hence it's abbreviated name.}

  TMHDR = record
    Toggles: byte; {Holds the initial status of Capslock, Numlock and ScrollLock}
    MKey : word;   {Macro hotkey}
    MacroEnd : Integer; {The number of messages recorded to the macro}
    MouseBase : TMouseBase;
    CodeHook : CodeHookProc; {A routine to call rather than play back keys}
    Name : Pchar;            {Macro name}
  end;

const
  MaxWord = 65520;
  {Even though memory is allocated for TMacroRec's based on the MaxMacroLen parameter
   passed into the InstallHook routine, type TMsgs is defined using the following
   constant so that the TEventMsgs that make up TMsgs can be viewed using the debugger.
   This is necessary because the debugger looks at the type definition to determine
   the max array index of array variables.  That is,

   TMsgs = array[1..1] of TEventMsg;

   would work just fine for our purposes since the memory TMsgs uses is allocated dynamically
   anyway.  However, this convention would fool the debugger, therefore we use the following.
   The same convention is used below with pTMacros.}
  TMsgMax = (MaxWord-sizeof(TMHDR)) div sizeof(TEventMsg);

type
  TMsgs = array[0..pred(TMsgMax)] of TEventMsg;

  {The actual definition of a macro record}

  TMacroRec = record
    MHDR : TMHDR;
    Msgs : TMsgs;
  end;

  {I don't establish pointers to types that don't yet exist; I build a pointer to
   a type after I've established the type.  There is seldom a need for setting up the
   pointer first (a linked list would be one example), and doing it without a need
   for it has always seemed awkward to me.
   }

  pTMacroRec = ^TMacroRec;
const
  MaxMacs = MaxWord div sizeof(pTMacroRec);

type

  {The array of macro handles stored in each macro context}

  TMacros = array [0..pred(MaxMacs)] of pTMacroRec;
  pTMacros = ^TMacros;

  {A specialized TMessage used to send messages to the CallingWindow}

  TWMMacMsg = record
    Msg: Word;
    State: word;     {Used in MM_XXXXCHANGED messages}
    Macro: longint;  {Used only in the MM_MACROASSIGNED message}
    Result: longint;
  end;

 {The macro context type.  It establishes a little macro "world", if you will,
  that is made the _current_ "world" by searching through the hotkeys of the
  allocated contexts for a match to the key or mouse button just pressed.  If a
  context can be located, the variables that were passed to it when InstallHook
  was called are copied to global versions of the same and the engine is allowed
  to proceed with the playback operation.}

TMacroContext = record
  TaskHandle : THandle;
  Macros : pTMacros;
  CurrentMacroRec : pTMacroRec;
  {The following are passed in from the caller}
  CallingWindowHook : HWND;
  MaxMacroNum : word;
  MaxMacroLen : word;
  FastPlayback : boolean;
  TaskScope : THandle;
  LoMsgBound, HiMsgBound : Word;
  MouseBase : TMouseBase;
  CancelChecking : Boolean;
end;

pTMacroContext = ^TMacroContext;
implementation
end.

